import pandas as pd
import numpy as np
from pandera.typing import DataFrame


def remove_duplicated_shifts(shifts: DataFrame) -> DataFrame:
    """
    Remove or flag shifts that are/could be duplicates.

    Parameters
    ----------
    shifts : DataFrame
        Data frame containing all shifts from a game.

    Returns
    -------
    game_shifts : DataFrame
        Data frame of all shifts with a new column to indicate possible duplicate shifts.

    """
    # Copy to avoid changing in-place
    game_shifts = shifts.copy()

    # Create a new column to flag for potential to flag for potential extra players
    game_shifts["PotentialExtraPlayer"] = False

    # See if the team has cases where an extra player is tagged
    for team, team_shifts in game_shifts.groupby("TeamName"):
        # The shift started at the start/end of the game
        start_of_game = team_shifts["ShiftStart"] == 0
        end_of_game = team_shifts["ShiftEnd"].isin([1200, 2400, 3600])
        
        # The start time of the shift is not present among team end times
        extra_start_player = ~team_shifts["ShiftStart"].isin(team_shifts["ShiftEnd"])

        # The end time of the shift is not present among team start times
        extra_end_player = ~team_shifts["ShiftEnd"].isin(team_shifts["ShiftStart"])
        
        # The shift occurred in regulation
        regulation = team_shifts["PeriodNumber"] < 4

        # Players who played an "extra" shift which they did not actually play
        extra = team_shifts.loc[((extra_start_player & ~start_of_game) | 
                                 (extra_end_player   & ~end_of_game)) & 
                                regulation]
        
        if len(extra) > 0:
            # Specify a potential extra shift
            game_shifts.loc[extra.index, "PotentialExtraPlayer"] = True

    # Remove shifts of duration 0
    game_shifts.loc[game_shifts["Duration"] != 0]        

    # Duplicated end time
    dupe_end = game_shifts.loc[game_shifts.duplicated(["PlayerId", "ShiftEnd"], keep=False)].copy()
    
    # Sort by duration, keep only the shortest
    dupe_end.sort_values(["PlayerId", "ShiftEnd", "Duration"], inplace=True)

    # Remove the longest duplicate per group
    drop_end_times = dupe_end.duplicated(["PlayerId", "ShiftEnd"], keep="first")
    
    # Drop the duplicated rows from the data
    game_shifts = game_shifts.drop(drop_end_times[drop_end_times].index)
    
    # Duplicated start time
    dupe_start = game_shifts.loc[game_shifts.duplicated(["PlayerId", "ShiftStart"], keep=False)].copy()
    
    # Sort by duration, keep only the shortest
    dupe_start.sort_values(["PlayerId", "ShiftStart", "Duration"], inplace=True)

    # Remove the longest duplicate per group
    drop_start_times = dupe_start.duplicated(["PlayerId", "ShiftStart"], keep="first")
    
    # Drop the duplicated rows from the data
    game_shifts = game_shifts.drop(drop_start_times[drop_start_times].index)

    return game_shifts


def merge_json_pbp_and_shifts(json_game) -> DataFrame:
    """
    Combine play by play and shifts from the JSON data into one data frame.

    Parameters
    ----------
    json_game : GameJSON
        Object with information from the JSON representation of the game.

    Returns
    -------
    events_with_players : DataFrame
        Adjusted data frame with players added to the play by play data.

    """
    # Extract pbp and shifts
    pbp = json_game.get_event_data()
    shifts = json_game.get_shifts()
    
    # Remove duplicate game shifts
    game_shifts = remove_duplicated_shifts(shifts)
   
    # Copy to avoid changing in-place    
    events = pbp.copy()
       
    # Ensure GameId is an integer
    game_shifts["GameId"] = game_shifts["GameId"].astype(int)
    events["GameId"] = events["GameId"].astype(int)
            
    # Get the game shifts
    shifts = game_shifts.copy()
        
    # Determine which team played at home
    home_team = events.HomeTeamName.unique()[0]
    
    # Combine event data with shifts to get all available information
    event_shifts = events.merge(shifts, on=["GameId", "PeriodNumber"], how="outer")

    # Keep only players who were (most likely) on the ice during the event
    on_ice = event_shifts.loc[(event_shifts.TotalElapsedTime.between(event_shifts.ShiftStart, 
                                                                     event_shifts.ShiftEnd-1) &
                               event_shifts.EventType.isin(["PERIOD START", "FACEOFF"]) &
                               event_shifts.PeriodNumber.le(3)) |
                              (event_shifts.TotalElapsedTime.between(event_shifts.ShiftStart+1,
                                                                     event_shifts.ShiftEnd) &
                               ~event_shifts.EventType.isin(["PERIOD START", "FACEOFF"])) |
                              (event_shifts.TotalElapsedTime.between(event_shifts.ShiftStart, 
                                                                     event_shifts.ShiftEnd-1) &
                               event_shifts.EventType.isin(["PERIOD START", "FACEOFF"]) &
                               event_shifts.PeriodNumber.gt(3)) |
                              (event_shifts.TotalElapsedTime.between(event_shifts.ShiftStart, 
                                                                     event_shifts.ShiftEnd) &
                               event_shifts.EventType.isin(["GOAL"]) &
                               event_shifts.PeriodNumber.gt(3)) 
                              ].copy()

    # Combine all player ids for each event number into a string
    player_ids_on_ice = on_ice[["TeamName", "EventNumber", "PlayerId"]].astype(str).groupby(
        ["TeamName", "EventNumber"])["PlayerId"].agg(",".join).unstack().T

    # Determine if the first team is the home team
    first_is_home = player_ids_on_ice.iloc[:, 0].name == home_team

    # Players on the ice for first and second team, respectively
    f1 = player_ids_on_ice.iloc[:, 0].str.split(pat=",", expand=True)
    f2 = player_ids_on_ice.iloc[:, 1].str.split(pat=",", expand=True)

    # Rename columns
    if first_is_home:
        f1.columns = [f"HomePlayer{i}" for i in range(1, len(f1.columns)+1)]    
        f2.columns = [f"AwayPlayer{i}" for i in range(1, len(f2.columns)+1)]    
    else:
        f1.columns = [f"AwayPlayer{i}" for i in range(1, len(f1.columns)+1)]    
        f2.columns = [f"HomePlayer{i}" for i in range(1, len(f2.columns)+1)]    
        
    # Combine home and away players into one data frame
    home_away_players = pd.merge(f1.reset_index(), f2.reset_index())
    
    # Convert all columns to float
    home_away_players = home_away_players.astype(float)
    
    # Convert eventnumber to integer
    home_away_players["EventNumber"] = home_away_players["EventNumber"].astype(int)
    
    # Sort by event number
    home_away_players.sort_values("EventNumber", inplace=True)
    
    # Combine event data with home and away players in the wide format
    events_with_players = events.merge(home_away_players, on="EventNumber")
    
    # Find players that may be duplicates
    possible_duplicates = on_ice.loc[on_ice.PotentialExtraPlayer, 
                                     ["TeamName", "EventNumber", "PlayerId",
                                      "PotentialExtraPlayer"]]
    
    # Loop over all shifts that maybe be erroneous
    for possible_duplicate in possible_duplicates.itertuples():
        
        # If it was the home team with the duplicate
        is_home = possible_duplicate.TeamName == home_team
        
        # See if there have been tagged with more than 6 players on the ice simultaneously
        cond = (events_with_players.EventNumber.eq(possible_duplicate.EventNumber) &
                (is_home and "HomePlayer7" in events_with_players.columns) |
                (~is_home and "AwayPlayer7" in events_with_players.columns))
        
        # Where the player is on the ice
        player = np.where(events_with_players.iloc[:, 30:].values == 
                          possible_duplicate.PlayerId)
        
        # In case there is no actual duplicate
        if len(player[0]) == 0:
            continue
        
        # Find the common index
        duplicate_idx = np.intersect1d(events_with_players.loc[cond].index, player[0])
        
        # Find the column where the index are duplicated
        duplicate_col = player[1][np.in1d(player[0], duplicate_idx)]
        
        for dupe_idx, dupe_col in zip(duplicate_idx, duplicate_col):
            # Find the column name of the duplicate
            col_name = events_with_players.columns[30+dupe_col]
            
            # Replace the duplicate player with nan
            events_with_players.loc[dupe_idx, col_name] = np.nan
    
    # Shift the home player columns to the left to fill NA holes        
    home = events_with_players.loc[:, events_with_players.columns.str.startswith("HomePlayer")].apply(
        lambda x: pd.Series(x.dropna().to_numpy()), axis=1)

    # Save the shifted player values
    events_with_players.loc[:, [f"HomePlayer{i}" for i in range(1, len(home.columns)+1)]] = home.values

    # Shift the away player columns to the left to fill NA holes
    away = events_with_players.loc[:, events_with_players.columns.str.startswith("AwayPlayer")].apply(
        lambda x: pd.Series(x.dropna().to_numpy()), axis=1)
    
    # Save the shifted player values
    events_with_players.loc[:, [f"AwayPlayer{i}" for i in range(1, len(away.columns)+1)]] = away.values
    
    for team in ["Home", "Away"]:
        # Find if there are any columns that are extra, i.e., PlayerX, where X = 7/8/9
        player_cond = (events_with_players.columns.str.startswith(f"{team}Player") &
                       events_with_players.columns.str.contains("7|8|9|10|11|12"))
        # If there are no extra columns
        if all(~player_cond):
            continue
        
        # Get the data corresponding to extra columns
        player_extra_columns = events_with_players.loc[:, player_cond]
        
        # Find rows that might have duplicated player ids
        player_duplicate = np.where(player_extra_columns.notna())
        
        # For the players who are duplicated, set the PlayerX to nan
        events_with_players.loc[player_duplicate[0], 
                                player_extra_columns.columns[player_duplicate[1]]] = np.nan
        
        # Remove duplicated columns
        events_with_players = events_with_players.loc[:, ~events_with_players.columns.duplicated()].copy()
        
    # Find columns with only NA values
    only_na_columns = events_with_players.notna().sum(axis=0)
    
    # Remove columns with only NA
    events_with_players = events_with_players.loc[:, only_na_columns.ne(0)].copy()
    
    # Add penalty shootout events too not drop any events
    events_with_players = pd.concat([
        events_with_players, 
        events.loc[events.TotalElapsedTime.eq(4800)]]).reset_index(drop=True)

    return events_with_players


def merge_json_and_html_pbp(json_game, html_game) -> DataFrame:
    """
    Combine play by play data from the JSON and HTML representations.

    Parameters
    ----------
    json_game : GameJSON
        Object containing the JSON representation of the game.
    html_game : GameHTML
        Object containing the HTML representation of the game.

    Returns
    -------
    html_pbp : DataFrame
        Merged data frame with players from the HTML source.

    """
    # Extract the play by play data
    json_pbp = json_game.pbp.copy()
    html_pbp = html_game.pbp.copy()

    # Drop the events that are not important
    event_drops = ["DELPEN", "SOC", "ANTHEM", "PGEND", "PGSTR", "EGT"]
    
    # Remove unwanted events
    html_pbp = html_pbp.loc[~html_pbp.EventType.isin(event_drops)].copy()
    json_pbp = json_pbp.loc[~json_pbp.EventType.isin(event_drops)].copy()
        
    # If the team of the event was the home team
    home_team = html_pbp.Team.eq(html_pbp.HomeTeamName)
    
    # Find all blocked shots in the HTML data
    blocks = html_pbp.EventType.eq("BLOCKED SHOT")
    
    # Flip the perspective of the blocked shots
    html_pbp.loc[blocks, ["Player1", "Player2"]] = html_pbp.loc[blocks, ["Player2", "Player1"]].values
    html_pbp.loc[blocks, ["PlayerId1", "PlayerId2"]] = html_pbp.loc[blocks, ["PlayerId2", "PlayerId1"]].values
    html_pbp.loc[blocks, ["PlayerType1", "PlayerType2"]] = html_pbp.loc[blocks, ["PlayerType2", "PlayerType1"]].values
    
    # If the home team blocked the shot; set team name to away
    html_pbp.loc[blocks & home_team, "Team"] = html_pbp.loc[blocks & home_team, "AwayTeamName"]
    
    # If the away team blocked the shot; set team name to home
    html_pbp.loc[blocks & ~home_team, "Team"] = html_pbp.loc[blocks & ~home_team, "HomeTeamName"]
    
    # Ensure GameId is an integer
    html_pbp["GameId"] = html_pbp["GameId"].astype(int)
    json_pbp["GameId"] = json_pbp["GameId"].astype(int)
    
    # Specify the columns to add
    cols = ["GoalsAgainst", "GoalsFor", "X", "Y", "GameWinningGoal", "EmptyNet"]
    
    # Find the index of the zone column
    zone_col_idx = int(np.where(html_pbp.columns.str.contains("Zone"))[0][0])
    
    # Loop over each column
    for col in cols[::-1]:
        # Insert the new column at the specified index
        html_pbp.insert(zone_col_idx, col, np.nan)
        
    # Loop over all the events in each game
    for game_id, events in html_pbp.groupby("GameId"):
        # Get all events from the JSON data of the same game
        json_events = json_pbp.loc[json_pbp.GameId.eq(game_id)].copy()
        
        # If there are no JSON events to add
        if len(json_events) == 0:
            break
        
        # Loop over each event
        for event in events.itertuples():
            # Get the same event type from the JSON data
            same_json_event_type = json_events.loc[json_events.EventType.eq(event.EventType) &
                                                   json_events.PeriodNumber.eq(event.PeriodNumber) &
                                                   (np.isnan(event.PenaltyMinutes) | 
                                                    (~np.isnan(event.PenaltyMinutes) &
                                                     json_events.PenaltyMinutes.eq(event.PenaltyMinutes))
                                                    )].copy()
            
            # Compute the time difference between the current event and the other events
            time_diff = same_json_event_type.TotalElapsedTime - event.TotalElapsedTime
            
            # If no close difference is found
            if all(time_diff > 30):
                continue
            
            # Find the JSON event with the closest time
            min_time_diff = min(time_diff)
            closest_time_idx = [idx for idx, diff in enumerate(time_diff) if diff == min_time_diff]
            
            # Get the event with the closest time
            closest_time = same_json_event_type.iloc[closest_time_idx]
            
            # print(len(closest_time))
            if len(closest_time) > 1:
                # Get the current length
                curr_len = len(closest_time)
                
                if event.PlayerId1 is not None:
                    # Find the player involved with the event
                    closest_time = closest_time.loc[closest_time.PlayerId1.eq(event.PlayerId1) |
                                                    np.isnan(event.PlayerId1)].copy()
                
                if len(closest_time) > 1:
                    # If there are still more than one match, use the first one
                    closest_time = closest_time.iloc[[0], :].copy()
                elif len(closest_time) == 0 and curr_len > 1:
                    # Get the first match if many still remain
                    closest_time = same_json_event_type.iloc[closest_time_idx].iloc[[0]]
            
            # Save the event in the HTML data
            html_pbp.loc[event.Index, cols] = closest_time[cols].values[0]
            
            # Remove the observation from the JSON data
            json_events.drop(closest_time.index, inplace=True)
    
    ## Update goals against / goals for
    # Get all goals
    goals = html_pbp.loc[html_pbp.EventType.eq("GOAL")]
    
    # Determine if the home team scored
    home_goal = goals.HomeTeamName.eq(goals.Team)
    
    # Current index 
    curr_idx = 0
    
    # Current number of home and away goals
    curr_home_goals = 0
    curr_away_goals = 0
    
    # Get the index of all goals
    index_of_goals = set(goals.index)
    
    # Get the index of the last event
    max_idx = html_pbp.index.max()
    
    # Add the index of the last event in case it is not already present
    index_of_goals.add((max_idx))
    
    # Loop over each goal index
    for goal_idx in sorted(index_of_goals):
        # Update the goals scored for all events prior
        html_pbp.loc[curr_idx:goal_idx, "GoalsAgainst"] = curr_away_goals
        html_pbp.loc[curr_idx:goal_idx, "GoalsFor"] = curr_home_goals
        
        # Update the score
        if goal_idx != max_idx and home_goal[goal_idx]:
            curr_home_goals += 1
        elif goal_idx != max_idx and ~home_goal[goal_idx]:
            curr_away_goals += 1
        
        # Update the goals scored for the actual goal
        html_pbp.loc[goal_idx, "GoalsAgainst"] = curr_away_goals
        html_pbp.loc[goal_idx, "GoalsFor"] = curr_home_goals
            
        # Update the current index
        curr_idx = goal_idx
    
    # Adjust goals against/for to only count after the goal has been confirmed
    html_pbp.loc[html_pbp.AwayTeamName.eq(html_pbp.Team) & 
                 html_pbp.EventType.eq("GOAL") &
                 html_pbp.PeriodNumber.le(3), "GoalsAgainst"] -= 1
    
    html_pbp.loc[html_pbp.HomeTeamName.eq(html_pbp.Team) & 
                 html_pbp.EventType.eq("GOAL") &
                 html_pbp.PeriodNumber.le(3), "GoalsFor"] -= 1
    
    # Fill the values for the events that occur after the last goal
    goal_cols = ["GoalsAgainst", "GoalsFor"]
    html_pbp.loc[:, goal_cols] = html_pbp.loc[:, goal_cols].ffill()
    
    return html_pbp


def merge_json_and_html_shifts(json_shifts: DataFrame, html_shifts: DataFrame) -> DataFrame:
    """
    Combine shifts from the JSON and HTML representations.

    Parameters
    ----------
    json_shifts : DataFrame
        Shifts from the JSON representation of the game.
    html_shifts : DataFrame
        Shifts from the HTML representation of the game.

    Returns
    -------
    shifts : DataFrame
        Merged data frame with shifts from both sources.

    """

    # Sort shifts by player id and shift start and end time
    sorted_json_shifts = json_shifts.sort_values(["PlayerId", "ShiftStart", "ShiftEnd"]
                                                 ).reset_index(drop=True).copy()
    
    sorted_html_shifts = html_shifts.sort_values(["PlayerId", "ShiftStart", "ShiftEnd"]
                                                 ).reset_index(drop=True).copy()
        
    # Loop over all end/start of periods from Period 1/2 to 11/12
    for time_idx, time in enumerate(range(1200, 12001, 1200)):
        # Find the shifts that overlap between periods
        json_shift_overlap_bool = (sorted_json_shifts.ShiftStart.lt(time) & 
                                   sorted_json_shifts.ShiftEnd.gt(time))
        html_shift_overlap_bool = (sorted_html_shifts.ShiftStart.lt(time) & 
                                   sorted_html_shifts.ShiftEnd.gt(time))
        
        # Get the overlapping shifts
        json_shift_overlap = sorted_json_shifts.loc[json_shift_overlap_bool].copy()
        html_shift_overlap = sorted_html_shifts.loc[html_shift_overlap_bool].copy()
        
        # Update the end time to be the end of the period
        sorted_json_shifts.loc[json_shift_overlap.index, "ShiftEnd"] = time
        sorted_html_shifts.loc[html_shift_overlap.index, "ShiftEnd"] = time
        
        # Update the period and time of the shift start
        json_shift_overlap[["PeriodNumber", "ShiftStart"]] = [time_idx + 2, time]
        html_shift_overlap[["PeriodNumber", "ShiftStart"]] = [time_idx + 2, time]
        
        # Add to the data 
        sorted_json_shifts = pd.concat([sorted_json_shifts, json_shift_overlap])
        sorted_html_shifts = pd.concat([sorted_html_shifts, html_shift_overlap])
        
    # Sort values and reset index
    sorted_json_shifts = sorted_json_shifts.sort_values(
        ["PlayerId", "ShiftStart", "ShiftEnd"]).reset_index(drop=True)
    
    sorted_html_shifts = sorted_html_shifts.sort_values(
        ["PlayerId", "ShiftStart", "ShiftEnd"]).reset_index(drop=True)
    
    # Update the duration to be correct
    sorted_json_shifts["Duration"] = sorted_json_shifts["ShiftEnd"] - sorted_json_shifts["ShiftStart"]
    sorted_html_shifts["Duration"] = sorted_html_shifts["ShiftEnd"] - sorted_html_shifts["ShiftStart"]

    # Ensure both columns are the correct data type
    sorted_html_shifts = sorted_html_shifts.astype({"GameId": int, "PlayerId": int})
    sorted_html_shifts = sorted_html_shifts.astype({"GameId": int, "PlayerId": int})

    # Combine both data sources into one dataframe
    shifts = sorted_html_shifts.drop(["Player", "ShiftNumber"], axis=1).merge(
        sorted_json_shifts.drop(["ShiftNumber"], axis=1), how="outer", indicator="Source")
        
    # Adjust the merge indicator column
    shifts["Source"].replace({"both": "Both", "left_only": "HTML", "right_only": "JSON"}, inplace=True)
    
    # Create a dictionary of player names and ids in the HTML shift data
    html_player_names = sorted_html_shifts[["Player", "PlayerId"]].set_index("PlayerId").to_dict()["Player"]
    
    # Find the id of any missing player names
    missing_player_names_id = shifts.loc[shifts.Player.isna(), "PlayerId"].values
    
    # Find the name of the missing players
    missing_player_names = {player_id: html_player_names[player_id].title() for 
                            player_id in missing_player_names_id}
    
    # Add any missing player names back
    shifts.loc[shifts.Player.isna(), "Player"] = shifts.loc[shifts.Player.isna(), "PlayerId"].replace(
        missing_player_names)

    # Determine the shift number by player
    shift_number = shifts.groupby("PlayerId")["ShiftStart"].rank(method="first").astype(int)
    
    # Add shift number back as a column
    shifts.insert(3, "ShiftNumber", shift_number)
    
    # Move the player column 
    player = shifts.pop("Player")
    shifts.insert(2, "Player", player)
    
    # Fill any NA with "" for Event
    shifts["Event"] = shifts.Event.fillna("")
    
    return shifts


import requests
from typing import Union, List, Tuple


def get_number_of_games_in_season(season: Union[str, int], season_type: str="R", 
                                  requests_session: requests.sessions.Session=None) -> Tuple[int, str]:
    """
    Retrieve the number of games for 

    Parameters
    ----------
    season : int or str
        Integer denoting the starting year of the season, e.g., 2021 gives 2021-2022.
    season_type : str
        One of: R = Regular Season, P = Playoffs, PR = Preseason, A = All-star.
    requests_session : requests.sessions.Session
        Session for requests.

    Returns
    -------
    n_games : int
        The total number of games in the season.

    """    
    # Create a new session if none is provided
    if requests_session is None:
        requests_session = requests.Session()
    
    # Converting the season type to its numerical representation
    season_type_code_dict = {"PR": "01", "R": "02", "P": "03", "A": "04"}    
    
    # Convert the season to an integer
    year = int(season)
    
    # Url for the season
    season_url = f"https://statsapi.web.nhl.com/api/v1/schedule?season={year}{year+1}&gameType={season_type}"
            
    # Number of games played during the season
    n_games = requests_session.get(season_url).json()["totalGames"]
        
    # Get the season type code
    season_type_code = season_type_code_dict[season_type]
    
    return n_games, season_type_code


def get_game_ids_between_dates(start_date: str, end_date: str,
                               requests_session: requests.sessions.Session=None) -> List[int]:
    """
    Find all game ids that took place between start and end date. 

    Parameters
    ----------
    start_date : str
        Date in YYYY-MM-DD format.
    end_date : str
        Date in YYYY-MM-DD format.
    requests_session : requests.sessions.Session
        Session for requests.
        
    Returns
    -------
    date_game_ids : list
        List of all game ids, given as integers, between the two dates.

    """
    # Create a new session if none is provided
    if requests_session is None:
        requests_session = requests.Session()
        
    # Get the information of all games between the two dates
    date_url = f"https://statsapi.web.nhl.com/api/v1/schedule?startDate={start_date}&endDate={end_date}"
    
    # Get the game information between the two dates
    games_between_dates = requests.get(date_url).json()["dates"]
    
    # Get all games for a given date
    date_games = [date["games"] for date in games_between_dates]
    
    # Find all game ids between the two dates
    date_game_ids = [game["gamePk"] for date in date_games for game in date]

    return date_game_ids


import numpy as np
import pandas as pd
import re
from itertools import chain
from datetime import datetime
import requests
from selectolax.parser import HTMLParser
from pandera.typing import DataFrame, Series
from typing import Tuple, List


class GameHTML:
    """
    Extract information from the HTML report of an NHL game.
    
    Arguments:
        game_id: int, the id of the game, given as e.g. 2021020001.
        requests_session: optional, the session to use for scraping.
    
    Attributes:
        game_id: int, the id of the game.
        away_team: str, the name of the away team.
        home_team: str, the name of the home team.
        url: str, the url of the HTML report. One for play-by-play, and one for each home and away shifts.
        tree: selectolax.parser.HTMLParser, parser for the page content. One for each url.
        event_base: list, list of all nodes (from the tree) for each event in the game.
    
    Methods:
        get_event_data: retrieves the event data from the game.
        get_shifts: retrieves all shifts from the home or away team.
    
    """
    def __init__(self, game_id: int, pbp: bool=True, shifts: bool=True,
                 requests_session: requests.sessions.Session=None):
        """ Initialize a new game object. """
        
        # Create a new session if none is provided
        if requests_session is None:
            requests_session = requests.Session()
        
        # Save the game id as attribute
        self.game_id = game_id

        # Save the urls
        self.pbp_url = self.__get_url(game_id, report_type="PL")
        self.home_shifts_url = self.__get_url(game_id, report_type="TH")
        self.away_shifts_url = self.__get_url(game_id, report_type="TV")

        # Save the parsed HTML tree
        self.pbp_tree = self.__parse_url(requests_session, report_type="PL")
        self.home_shifts_tree = self.__parse_url(requests_session, report_type="TH")
        self.away_shifts_tree = self.__parse_url(requests_session, report_type="TV")

        # Extract the event base from the HTML
        self.event_base = self.__get_event_base()

        # Get team names
        team_names = self.__get_team_names()

        # Add home and away team names    
        self.away_team = team_names[0]
        self.home_team = team_names[1]

        # Get and set the game info
        self.game_date = self.__get_game_info()

        
    def __get_url(self, game_id: int, report_type: str="PL") -> str:
        """ Get the url used for retrieving data. """
        
        if report_type not in ["PL", "TH", "TV"]:
            raise ValueError("The report type is not supported.")
        
        # Extract season
        season = int(f"{game_id}"[:4])

        # Convert the season name to season-season+1
        season_full = f"{season}{season+1}"

        # Extract the unique identifying features of the game
        game_id_short = f"{game_id}"[5:]

        # Specify the url
        url = f"https://www.nhl.com/scores/htmlreports/{season_full}/{report_type}0{game_id_short}.HTM"

        return url


    def __parse_url(self, requests_session: requests.sessions.Session, report_type: str="PL") -> HTMLParser:
        """ Parse the url."""
        
        if report_type == "PL":
            url = self.pbp_url
        elif report_type == "TH":
            url = self.home_shifts_url
        elif report_type == "TV":
            url = self.away_shifts_url
        else:
            raise ValueError("The report type is not supported.")
        
        # Retrieve the page
        page = requests_session.get(url)

        if page.status_code == 404:
            raise ValueError("The webpage does not exist.")

        # Parse the content on the page
        tree = HTMLParser(page.content)

        return tree


    def __get_team_names(self) -> List:
        """ Get the away and home teams in the game. """
        # Get the nodes satisfying the condition
        parsed_tree = self.pbp_tree.css("tr > td[class^=heading]")
        
        # Get all teams in the game
        teams = [node.text() for node in parsed_tree if "On Ice" in node.text()][:2]
        
        # Keep only the team acronym
        teams = [team[:3] for team in teams]
        
        return teams


    def __get_game_info(self) -> Tuple[int, str]:
        """ Get the about-game information from the game. """
        
        # Get the game info
        game_info = self.pbp_tree.css("table[id=GameInfo]")[0].text()

        # Remove whitespace
        game_info = re.sub("\xa0|\n", " ", game_info)

        # Remove unwanted info
        game_info = re.sub("\s+Play By Play\s+|Start[\s\S]+", "", game_info)

        # Get the date of the game (as a string)
        try:
            game_date = re.findall("(?<=,)[\s\S]+(?=Attendance)", game_info)[0].strip()
        except IndexError:
            game_date = np.nan

        try:
            # Convert the game from a string to an integer
            game_date_int = datetime.strftime(datetime.strptime(game_date, "%B %d, %Y"), "%Y%m%d")
        except TypeError:
            game_date_int = 0
        
        return game_date_int


    def __get_event_base(self) -> List:
        """ Get list of all nodes with all events, i.e., starts with id=PL. """
        
        # Get the event base from the newest seasons
        event_base = self.pbp_tree.css("tr[id^=PL]")
        
        if int(str(self.game_id)[:4]) < 2021 or len(event_base) == 0:
            # Search for color rather than id if id does not exist
            event_base = self.pbp_tree.css("tr[class$=Color]")
                       
            # If there are not id given for each node, create one.
            for idx, event in enumerate(event_base):
                event.attrs["id"] = f"PL-{idx+1}"
        
        return event_base


    def _get_event_number(self) -> np.ndarray:
        """ Get the event number from the event nodes. """
        
        # Extract event number
        event_number_list = [int(re.sub("PL-", "", event.id))
                             if bool(re.search("(?<=\n)\d+(?=\n\d)", event.text())) else 1
                             for event in self.event_base]

        return np.array(event_number_list)


    def _get_period_number(self) -> np.ndarray:
        """ Get the period number from the event nodes. """
        
        # Extract period number
        period_number_list = [int(re.findall("(?<=\d\n)\d(?=\n)", event.text())[0])
                              if bool(re.search("(?<=\d\n)\d(?=\n)", event.text())) else 1
                              for event in self.event_base]

        return np.array(period_number_list)


    def _get_event_manpower(self) -> np.ndarray:
        """ Get the manpower from the event nodes."""
        
        # Extract manpower situation
        event_manpower_list = [re.findall("(?<=\d\n)[A-z]+(?=\n\d)", event.text())
                               for event in self.event_base]
        
        # Convert to string representation
        event_manpower_list = [event[0] if len(event) > 0 else np.nan 
                               for event in event_manpower_list]

        return np.array(event_manpower_list, dtype=object)


    def _get_event_time(self) -> np.ndarray:
        """ Get the event time from the event nodes."""
        
        # Extract the time of each event
        event_time_list = [re.findall("(?<=[A-z]\n)\d+:\d{2}(?=\d)|(?<=\xa0\n)\d+:\d{2}(?=\d)", 
                                      event.text())
                           for event in self.event_base]
        
        # Convert to string representation
        event_time_list = [event[0] if len(event) > 0 else np.nan 
                           for event in event_time_list]

        return np.array(event_time_list, dtype=object)


    def _get_event_type(self) -> np.ndarray:
        """ Get the event type from the event nodes."""
        
        # Extract the type of event
        event_type_list = [re.findall("(?<=:\d{2}\n)[A-z]+(?=\n)", event.text())
                           for event in self.event_base]

        # Convert to string representation
        event_type_list = [event[0] if len(event) > 0 else np.nan 
                           for event in event_type_list]

        return np.array(event_type_list, dtype=object)


    def _get_event_description(self) -> np.ndarray:
        """ Get the description from the event nodes. """
        
        # Extract the description of each event
        event_description_list = [re.findall("(?<=[A-z]\n)[A-z]+.+(?=\n)", 
                                             re.sub("\xa0", " ", event.text()))
                                  for event in self.event_base]

        # Convert to string representation
        event_description_list = [event[0] if len(event) > 0 else np.nan 
                                  for event in event_description_list]

        return np.array(event_description_list, dtype=object)


    def _get_event_team(self, event_description_list: np.ndarray) -> Series:
        """ Get the team of the event from the event description."""
        
        # Get only the first team of the event
        event_team_list = [re.findall("(?<=^)[A-Z]{3}(?=\s)|(?<=^)[A-Z]\.[A-Z](?=\s)", 
                                       re.sub("HIT", "", description))
                            if not isinstance(description, float) else np.nan
                            for description in event_description_list]
        
        # Convert to string representation
        event_team_list = [event[0] if not isinstance(event, float) and 
                            len(event) > 0 else np.nan 
                            for event in event_team_list]
        
        return pd.Series(event_team_list, name="Team")
    
    
    def _get_event_players(self, event_description_list: np.ndarray) -> Series:
        """ Get the players involved in the event from the event description."""
    
        # Get all players who were involved
        event_players_list = [re.findall("\#\d+ [\w+'-]+", 
                                         re.sub("\.", "", description))
                              if not isinstance(description, float) else np.nan
                              for description in event_description_list]
        
        return pd.Series(event_players_list, name="Players")


    def _get_event_zone(self, event_description_list: np.ndarray) -> Series:
        """ Get the zone from the event description."""
    
        # Get the zone of the event
        event_zone_list = [re.findall("[A-z]{3}(?=\. Zone)", description)
                              if not isinstance(description, float) else np.nan
                              for description in event_description_list]
        
        # Convert to string representation
        event_zone_list = [event[0] if not isinstance(event, float) and 
                           len(event) > 0 else np.nan 
                           for event in event_zone_list]
        
        return pd.Series(event_zone_list, name="Zone")


    def _get_event_shot_type_and_distance(self, event_description_list: np.ndarray) -> Tuple[Series, Series]:
        """ Get the shot type and distance of the event from the event description."""
                
        # Get the shot type for all shots
        event_shot_type_list = [re.findall("(?<=, )[A-z]+(?=,)|(?<=, )[A-z]+-[A-z]+(?=,)", 
                                           re.sub("Goalpost", "", description))
                                if not isinstance(description, float) else np.nan
                                for description in event_description_list]
        
        # Get the shot distance
        event_shot_distance_list = [re.findall("\d+(?= ft.)", description)
                                    if not isinstance(description, float) else np.nan
                                    for description in event_description_list]
        
        # Convert to string representation
        event_shot_type_list = [event[0] if not isinstance(event, float) and 
                                len(event) > 0 else np.nan 
                                for event in event_shot_type_list]
        
        # Convert to numeric representation
        event_shot_distance_list = [float(event[0]) if not isinstance(event, float) and 
                                    len(event) > 0 else np.nan 
                                    for event in event_shot_distance_list]
        
        return (pd.Series(event_shot_type_list, name="ShotType"), 
                pd.Series(event_shot_distance_list, name="ShotDistance"))

    
    def _get_event_penalty(self, event_description_list: np.ndarray) -> Tuple[Series, Series]:
        """ Get the name of the penalty and its duration from the event description."""
        
        # Get all penalties, inlcuding type and severity
        event_penalty_list = [re.findall("(?<=[A-Z] ).+\(\d+ min\)(?=,)|(?<=[A-Z] ).+\(\d+ min\)",
                                         re.sub("\#\d+ [\w+'-]+| Served By:\s+| [A-Z]+(?=, [A-z]{3}\. Zone)", "",
                                                re.sub("\xa0", " ", event)))
                              if not isinstance(event, float) and len(event) > 0 else np.nan 
                              for event in event_description_list]
        
        # Get the name of the penalty type
        event_penalty_name_list = [re.sub("\([\s\S]+\)|[A-Z]+ (?=[A-Z][a-z]+)|\#", "", 
                                          event[0].strip()).strip()
                                   if not isinstance(event, float) and 
                                   len(event) > 0 else np.nan 
                                   for event in event_penalty_list]
        
        # Get the duration of the penalty
        event_penalty_duration_list = [re.sub("[\s\S]+\(|\)", "", event[0].strip()) 
                                       if not isinstance(event, float) and 
                                       len(event) > 0 else np.nan 
                                       for event in event_penalty_list]
        
        return (pd.Series(event_penalty_name_list, name="PenaltyType"), 
                pd.Series(event_penalty_duration_list, name="PenaltyMinutes"))


    def _penalty_shot_bool(self, event_description_list: np.ndarray) -> Series:
        """ Determine if there were any penalty shots from the event description."""
        
        # Get all penalties, inlcuding type and severity
        event_penalty_shot_list = [bool(re.search("Penalty Shot", event))
                                   if not isinstance(event, float) and len(event) > 0 else False
                                   for event in event_description_list]
        
        return pd.Series(event_penalty_shot_list, name="PenaltyShot")


    def _get_team_players(self, idx_to_drop: np.ndarray, home: bool=True) -> Tuple[np.ndarray, dict,
                                                                                   pd.Series]:
        """ Get the players from the home/away team on the ice. """
        
        # Clone the original tree
        tree_copy = self.pbp_tree.clone()
        
        # Loop over all ids to drop
        for idx in idx_to_drop:
            # Find the matching node
            for tag in tree_copy.css(f"tr[id=PL-{idx+1}]"):
                # Remove nodes that does not have players on the ice
                tag.decompose()
                
        # Get the event base of the new tree
        new_event_base = tree_copy.css("tr[id^=PL]")

        # Specify the team selector
        team_css_selector = "td[class$=bborder]" if home else "td[class$=rborder]"

        # Get all the nodes which may contain players from the team
        player_nodes = [event.css(team_css_selector)
                        for event in new_event_base]
        
        # Get all the nodes that contain players from the team
        player_nodes = [[node.css("font[style=\"cursor:hand;\"]") 
                         for node in node_list if 
                         node.css_matches("font[style=\"cursor:hand;\"]")][0]
                        for node_list in player_nodes]
        
        # Remove all players who do not have the attribute "title"
        players = [[player for player in player_list if 
                    "title" in player.attributes] for 
                   player_list in player_nodes]
                    
        # Get all player names and positions
        player_names = [[player.attributes["title"] if "title" in player.attributes
                         else re.findall("(?<=title=\")[\s\S]+(?=\">)", player.html)
                         for player in player_list] for
                        player_list in players]

        # Create a nested tuple of (# LAST NAME, FULL NAME)
        player_number_name = [[(f"#{player.text()} {player.attributes['title'].split(' - ')[1].split()[1].replace('.', '')}", 
                                f"{player.attributes['title'].split(' - ')[1]}")
                               if bool(re.search("[A-Z]", player.attributes["title"])) else None
                               for player in player_list ] for
                              player_list in players]
        
        # Remove any None values
        for list_idx, player_list in enumerate(player_number_name):
            if None in player_list:
                # Only keep non None values
                player_list_new = [player for player in player_list if player is not None]
                
                # Overwrite the pre-existing list
                player_number_name[list_idx] = player_list_new
        
        # Flatten the list of lists and convert to a set
        unique_player_number_name = set(chain.from_iterable(player_number_name))

        # Create a dictionary of mappings from # LAST NAME to FULL NAME
        number_name_map = {item[0]: item[1] for item in unique_player_number_name}

        # Get all player positions
        player_positions = [[player.split(" - ")[0] for player in player_list] for
                            player_list in player_names]

        # Get all player names
        player_names = [[player.split(" - ")[1] for player in player_list] for
                        player_list in player_names]

        # Check if the goalie was on the ice
        goalie_on_ice = ["Goalie" in player_list for player_list
                         in player_positions]
        
        # Find the goalie name for the team (if available)
        goalie_name = [np.array(name_list)[np.where([position == "Goalie" for position
                                                     in position_list])[0]]
                       for name_list, position_list in zip(player_names, player_positions)]
        
        # Convert to string representation
        goalie_name = [name[0] if len(name) > 0 else np.nan for name in goalie_name]
        
        # Determine the name of the series indicating if the goalie was on the ice
        side = "Home" if home else "Away"

        return (np.array(player_names, dtype=object), number_name_map, 
                pd.Series(goalie_name, name=f"{side}GoalieName"),
                pd.Series(goalie_on_ice, name=f"{side}GoalieOnIce"))

        
    def _expand_column_of_list(self, array: np.ndarray, column_name: str) -> DataFrame:
        """
        Expand a column that has list elements to multiple columns, each containing
        a unique element from the list. The columns are renamed as 'column_nameN' 
        where N is the number of columns 1, 2, ...
        """
        
        # Remove None if it exists
        column_series = pd.Series([[] if obj is None or isinstance(obj, float) 
                                   else obj for obj in array])
        
        # Expand the column of list to be one item per column
        column_df = pd.DataFrame(column_series.tolist())
        
        # Rename columns
        column_df.columns = [f"{column_name}{i}" for i in range(1, len(column_df.columns)+1)]
        
        return column_df


    def _define_player_roles(self, event_type_list: np.ndarray, players_involved: DataFrame) -> DataFrame:
        """ Define the player roles. """
        
        # Copy to avoid changing in-place
        players_involved = players_involved.copy()
        
        # Combine event and players involved
        event_type_series = pd.Series(event_type_list, name="EventType")        

        # Compute the number of players for each event
        nr_players_per_event = players_involved.notna().sum(axis=1)

        # Define the first player role
        players_involved["PlayerType1"] = np.select(
            [event_type_series.eq("FAC"),
             event_type_series.eq("HIT"),
             event_type_series.isin(["TAKE", "GIVE"]),
             event_type_series.isin(["SHOT", "BLOCK", "MISS"]),
             event_type_series.eq("GOAL"),
             event_type_series.eq("PENL")],
            ["Winner", "Hitter", "PlayerId", "Shooter", "Scorer", "PenaltyOn"],
            default=np.nan)

        # Define the second player role
        players_involved["PlayerType2"] = np.select(
            [event_type_series.eq("FAC")   & nr_players_per_event.eq(2),
             event_type_series.eq("HIT")   & nr_players_per_event.eq(2),
             event_type_series.eq("BLOCK") & nr_players_per_event.eq(2),
             event_type_series.eq("GOAL")  & nr_players_per_event.ge(2),
             event_type_series.eq("PENL")  & nr_players_per_event.ge(2)],
            ["Loser", "Hittee", "Blocker", "Assist", "DrewBy"],
            default=np.nan)
        
        # Define the third player role
        players_involved["PlayerType3"] = np.select(
            [event_type_series.eq("GOAL") & nr_players_per_event.eq(3),
             event_type_series.eq("PENL") & nr_players_per_event.eq(3)],
            ["Assist", "ServedBy"],
            default=np.nan)
        
        return players_involved


    def _map_team_tri_to_names(self) -> dict:
        """ Map a team's tri code to their full team name """
        
        # Create the mapping
        tri_to_team_map = {
            "ANA": "Anaheim Ducks",
            "ARI": "Arizona Coyotes",
            "ATL": "Atlanta Thrashers",
            "BOS": "Boston Bruins",
            "BUF": "Buffalo Sabres",
            "CGY": "Calgary Flames" ,
            "CAR": "Carolina Hurricanes",
            "CHI": "Chicago Blackhawks" ,
            "COL": "Colorado Avalanche",
            "CBJ": "Columbus Blue Jackets",
            "DAL": "Dallas Stars",
            "DET": "Detroit Red Wings",
            "EDM": "Edmonton Oilers" ,
            "FLA": "Florida Panthers",
            "L.A": "Los Angeles Kings",
            "LAK": "Los Angeles Kings",
            "MIN": "Minnesota Wild",
            "MTL": "Montréal Canadiens",
            "NSH": "Nashville Predators",
            "N.J": "New Jersey Devils",
            "NJD": "New Jersey Devils",
            "NYI": "New York Islanders",
            "NYR": "New York Rangers",
            "OTT": "Ottawa Senators",
            "PHI": "Philadelphia Flyers",
            "PHX": "Phoenix Coyotes",
            "PIT": "Pittsburgh Penguins",
            "S.J": "San Jose Sharks",
            "SJS": "San Jose Sharks",
            "SEA": "Seattle Kraken",
            "STL": "St. Louis Blues",
            "T.B": "Tampa Bay Lightning",
            "TBL": "Tampa Bay Lightning",
            "TOR": "Toronto Maple Leafs",
            "VAN": "Vancouver Canucks",
            "VGK": "Vegas Golden Knights",
            "WSH": "Washington Capitals",
            "WPG": "Winnipeg Jets"
            }
        
        return tri_to_team_map


    def _add_events_without_players_on_ice(self, idx_to_drop: np.ndarray, 
                                           event_number_list: np.ndarray,
                                           period_number_list: np.ndarray,
                                           event_manpower_list: np.ndarray,
                                           event_time_list: np.ndarray,
                                           event_type_list: np.ndarray,
                                           event_description_list: np.ndarray
                                           ) -> DataFrame:
        """ Add the events where there are no players on the ice but an 
            event description is still available. """
            
        # Find the entries which are not nan
        non_nan_entries = np.where([isinstance(event, str) for event in 
                                    event_description_list[idx_to_drop]])[0]
        
        # Get the indexes of events which have a description
        non_na_idx = idx_to_drop[non_nan_entries]
        
        # The events where the players are missing and an event description is available
        missing_players_df = pd.DataFrame([event_number_list[non_na_idx],
                                           period_number_list[non_na_idx],
                                           event_manpower_list[non_na_idx],
                                           event_time_list[non_na_idx],
                                           event_type_list[non_na_idx],                                                
                                           event_description_list[non_na_idx]],
                                          index=["EventNumber", "PeriodNumber", 
                                                 "Manpower", "EventTime", 
                                                 "EventType", "Description"]).T
        
        return missing_players_df


    def get_event_data(self) -> DataFrame:
        """ Get all events from the HTML RTSS report. """
        
        # Get the game metadata
        meta_data = (self.game_id, self.away_team, self.home_team, self.game_date)
        
        # Get the event and period numbers
        event_number_list = self._get_event_number()
        period_number_list = self._get_period_number()
        
        # Get the manpower situation
        event_manpower_list = self._get_event_manpower()
        
        # Get the time of the event
        event_time_list = self._get_event_time()
        
        # Get the type of event
        event_type_list = self._get_event_type()
        
        # Get the description of the event
        event_description_list = self._get_event_description()
        
        # Find the events to keep (i.e., events with players on ice)
        idx_to_keep = np.array([int(node.attributes["id"].split("-")[1]) for node in self.event_base if 
                                # If the away team has players for the given node
                                bool(re.search("bborder \+ rborder\">\n<table", node.html)) and
                                # If the home team has players for the given node
                                bool(re.search("bborder\">\n<table", node.html))])-1

        # Find the events to drop as the non-intersection
        idx_to_drop = np.setxor1d(event_number_list-1, idx_to_keep)
        
        # Get all events where players are missing but event description is available
        missing_players_df = self._add_events_without_players_on_ice(idx_to_drop, event_number_list,
                                                                     period_number_list, event_manpower_list,
                                                                     event_time_list, event_type_list,
                                                                     event_description_list)
        
        # Get player names and positions for each team
        away_player_names, away_name_map, away_goalie_name, away_goalie_on_ice = self._get_team_players(idx_to_drop, home=False)
        home_player_names, home_name_map, home_goalie_name, home_goalie_on_ice = self._get_team_players(idx_to_drop, home=True)
        
        # Combine the home and away name maps into one dictionary
        name_map = {**away_name_map, **home_name_map}
        
        # Expand a column of list to multiple columns for player names
        away_player_names_df = self._expand_column_of_list(away_player_names, "AwayPlayer")
        home_player_names_df = self._expand_column_of_list(home_player_names, "HomePlayer")

        # Combine all events into one data frame
        events = pd.DataFrame([event_number_list[idx_to_keep], period_number_list[idx_to_keep], 
                               event_manpower_list[idx_to_keep], event_time_list[idx_to_keep], 
                               event_type_list[idx_to_keep], event_description_list[idx_to_keep]],
                              index=["EventNumber", "PeriodNumber", 
                                     "Manpower", "EventTime", 
                                     "EventType", "Description"]).T               

        # Combine events and players on the ice
        events_with_players_on_ice = pd.concat([events,
                                                away_player_names_df.reset_index(drop=True), 
                                                home_player_names_df.reset_index(drop=True),
                                                away_goalie_on_ice, 
                                                away_goalie_name,
                                                home_goalie_on_ice,
                                                home_goalie_name],
                                               axis=1)
       
        # Add events with missing players
        events_and_missing = pd.concat([events_with_players_on_ice, missing_players_df]).sort_values(
            "EventNumber").reset_index(drop=True)
        
        # Get the zone of the event
        event_zone_list = self._get_event_zone(events_and_missing.Description)

        # Get the teams in the event
        event_team_list = self._get_event_team(events_and_missing.Description)
        
        # Get the players involved in the event
        event_players_list = self._get_event_players(events_and_missing.Description)
        
        # Get the shot type
        event_shot_type_list, event_shot_distance_list = self._get_event_shot_type_and_distance(events_and_missing.Description)
        
        # Get penalty information
        event_penalty_name_list, event_penalty_duration_list = self._get_event_penalty(events_and_missing.Description)

        # Determine if there were any penalty shots
        event_penalty_shot_bool = self._penalty_shot_bool(events_and_missing.Description)
        
        # Expand a column of list to multiple columns for team and involved players
        players_involved = self._expand_column_of_list(event_players_list, "Player")

        # Rename from # LAST NAME to FULL NAME
        players_involved.replace(name_map, inplace=True)

        # Define player roles
        players_involved = self._define_player_roles(events_and_missing.EventType, players_involved)
        
        # Combine events with all information regarding players, teams etc.
        events_with_all_info = pd.concat([events_and_missing, event_team_list, 
                                          players_involved, event_zone_list, 
                                          event_shot_type_list, event_penalty_shot_bool, 
                                          event_penalty_name_list, event_penalty_duration_list],
                                         axis=1)
        
        # Repeat meta data to match the length of the data
        meta_data_df = pd.DataFrame(np.repeat([meta_data], len(events_with_all_info), axis=0),
                                     columns=["GameId", "AwayTeamName", "HomeTeamName",
                                              "Date"])
        
        # Map team tri code to team name
        meta_data_df.AwayTeamName.replace(self._map_team_tri_to_names(), inplace=True)
        meta_data_df.HomeTeamName.replace(self._map_team_tri_to_names(), inplace=True)
        
        # Add meta data to the data frame
        pbp = pd.concat([meta_data_df, events_with_all_info], axis=1)
        
        # Map team to full name
        pbp.Team.replace(self._map_team_tri_to_names(), inplace=True)
             
        # Change data type to integer 
        pbp["Date"] = pbp.Date.astype(int)
        pbp["EventNumber"] = pbp.EventNumber.astype(int)
        pbp["PeriodNumber"] = pbp.PeriodNumber.astype(int)
                
        # Convert penalty minutes to an integer
        pbp["PenaltyMinutes"] = pbp.PenaltyMinutes.astype(str).str.replace(" min", "").astype(float)  
        
        # Reduce event number by 2 to account for the extra events before the game
        if pbp.iloc[0].EventType == "PGSTR":
            pbp["EventNumber"] -= 2
        
        # Adjust event number of later periods to account for additional events between periods
        for period_number in pbp.loc[pbp.PeriodNumber.gt(1)].PeriodNumber.unique():
            pbp.loc[pbp.PeriodNumber.eq(period_number), "EventNumber"] += 2 * (period_number - 1)
        
        # Find all delayed penalties and unusual events/double stoppage for a challenge
        unusual = (pbp.EventType.isin(["DELPEN", "EGT"]) |
                   (pbp.EventType.eq("STOP") & pbp.EventType.shift(-1).eq("CHL") &
                    pbp.EventType.shift(-2).eq("STOP")))
        
        if any(unusual) > 0:
            # Get the unusual events event numbers
            unusual_event_numbers = pbp.loc[unusual, "EventNumber"].values
            
            # Add the final event number to the array
            unusual_event_numbers = np.concatenate((unusual_event_numbers, 
                                                    np.array([pbp.EventNumber.max()])))
            
            for idx, event_number in enumerate(unusual_event_numbers[1:]):
                # Get the event number of the previous unsual event
                previous_event_number = unusual_event_numbers[idx]
                
                # Find the event number that need to be adjusted
                event_number_to_change = (pbp.EventNumber.gt(previous_event_number) &
                                          pbp.EventNumber.le(event_number))
                
                # Reduce the event number by 1
                pbp.loc[event_number_to_change, "EventNumber"] -= 1 * (idx + 1)
            
            # Remove delayed penalties and unusual events
            pbp = pbp.loc[~unusual].copy()
        
        # All penalties served by another player
        serving_penalties = (pbp.EventType.eq("PENL") &
                             pbp.Description.astype(str).str.contains("Served By"))
        
        # Add missing column if needed
        if "Player3" not in pbp.columns:
            pbp["Player3"] = np.nan
            pbp["PlayerType3"] = np.nan
            
        # Switch player names as they are incorrect when a third player serves the penalty
        pbp.loc[serving_penalties & pbp.PlayerType3.ne("nan"), 
                       ["Player3", "Player2"]] = \
        pbp.loc[serving_penalties & pbp.PlayerType3.ne("nan"), 
                       ["Player2", "Player3"]].values
        
        # For players serving the penalty of a goalie
        pbp.loc[serving_penalties & pbp.PlayerType3.eq("nan") & 
                pbp.PlayerType2.ne("nan"), "PlayerType2"] = "ServedBy"
        
        # For players serving a team penalty
        pbp.loc[serving_penalties & pbp.PlayerType3.eq("nan") & 
                pbp.PlayerType2.eq("nan"), "PlayerType1"] = "ServedBy"
            
        # Determine the winner of faceoffs
        faceoff_winning_team = pbp.loc[ pbp.EventType.eq("FAC"), "Description"].str.extract(
                "(^[A-Z]{3})", expand=False).replace(self._map_team_tri_to_names())
                
        # Switch the order of players if the home team won the faceoff
        pbp.loc[faceoff_winning_team.eq(pbp.HomeTeamName),
                       ["Player1", "Player2"]] = \
        pbp.loc[faceoff_winning_team.eq(pbp.HomeTeamName),
                       ["Player2", "Player1"]].values
        
        # Find all the columns with players on the ice
        player_columns = pbp.columns[pbp.columns.str.contains(
            "[HomeAway]Goalie|[HomeAway]Player\d")]
        
        # Reorder the columns in the data to have players on the ice at the end
        pbp = pd.concat([pbp.drop(player_columns, axis=1), 
                                pbp[player_columns]], axis=1)

        # Find columns that should not exist/are erroneous
        erroneous_cols = pbp.columns[pbp.columns.str.contains("\d{2}|^Player4", regex=True)]
        
        # Remove the erroneous columns
        pbp.drop(erroneous_cols, axis=1, inplace=True)
        
        # Fill NA time values with a placeholder
        pbp.EventTime.fillna("20:00", inplace=True)
        
        # Add total elapsed time
        pbp["TotalElapsedTime"] = 1200 * (pbp["PeriodNumber"] - 1) + [
            60 * int(x[0]) + int(x[1]) for x in pbp["EventTime"].str.split(":")
            ]

        # Rename events
        pbp.EventType.replace({
            "FAC": "FACEOFF", "STOP": "STOPPAGE", "CHL": "CHALLENGE",
            "GEND": "GAME END", "PSTR": "PERIOD START", "PEND": "PERIOD END",
            "GIVE": "GIVEAWAY", "TAKE": "TAKEAWAY", "PENL": "PENALTY",
            "MISS": "MISSED SHOT", "BLOCK": "BLOCKED SHOT"}, inplace=True)
    
        return pbp


    def get_shifts(self, home: bool=True):
        """ Get all shifts from the given game. """
        
        # Determine the team perspective
        if home: 
            tree = self.home_shifts_tree
        else:
            tree = self.away_shifts_tree
            
        # Get all nodes with player number and names
        player_names_css = tree.css("tr > td[class~=playerHeading]")
        
        # Extract player names
        player_names = [re.sub("\d+ ", "", re.findall("(?<=\>).+(?=\</)", player.html)[0]) 
                        for player in player_names_css]
        
        # Switch order of first and last names
        player_names = [f"{re.findall('(?<=, ).+', player)[0]} {re.findall('.+(?=, )', player)[0]}" if
                        len(re.sub("[\s,]", "", player)) > 0 else "Unknown"
                        for player in player_names]

        # Find all shifts
        shifts = tree.css("tr[class$=Color]")
        
        # Extract shift information for each shift
        shift_info = [shift.css("[class$=border]") for shift in shifts]
        
        # Extract the text from the shift information
        shift_info_text = [[re.findall("(?<=\>).+(?=\</)", s.html) for s in shift] for shift in shift_info]
        
        # Simplify the nest list structure
        shift_info_text = [[attr[0] if len(attr) > 0 else [] for attr in shift] for shift in shift_info_text] 
        
        # Create a data frame of the shifts
        shifts = pd.DataFrame(shift_info_text, 
                              columns=["ShiftNumber", "PeriodNumber",
                                       "ShiftStart", "ShiftEnd", 
                                       "Duration", "Event"]).dropna().reset_index(drop=True)
        
        # Replace "OT" with a 4 for period number and 5 for SO
        shifts["PeriodNumber"] = shifts.PeriodNumber.replace({"OT": 4, "SO": 5})
        
        # Replace any other non-numeric characters with zero
        shifts["PeriodNumber"] = shifts["PeriodNumber"].replace({"\D+": "0"}, regex=True)
        
        # Change column types
        cols = ["ShiftNumber", "PeriodNumber"]
        shifts[cols] = shifts[cols].astype(int, errors="ignore")
        
        # Create a grouping to indiciate unique players
        shifts["PlayerNr"] = (shifts.ShiftNumber != shifts.ShiftNumber.shift() + 1).cumsum()
        
        # Create a mapping from plyer nr to player name
        player_mapping = {nr: player for nr, player in zip(shifts["PlayerNr"].unique(), player_names)}
        
        # Insert the new column with player name
        shifts.insert(0, "Player", shifts["PlayerNr"].replace(player_mapping))   
        
        # Remove unwanted text
        shifts.replace("&nbsp;|\s/\s.+", "", regex=True, inplace=True)
        
        # Convert time columns to seconds
        shifts["ShiftStart"] = [int(minute) * 60 + int(second) for minute, second in 
                                shifts["ShiftStart"].str.split(":")]
        shifts["ShiftEnd"] = [int(minute) * 60 + int(second) for minute, second in 
                              shifts["ShiftEnd"].str.split(":")]
        shifts["Duration"] = [int(minute) * 60 + int(second) for minute, second in 
                              shifts["Duration"].str.split(":")]   
        
        # Add period number offsets
        shifts["ShiftStart"] = (1200 * (shifts["PeriodNumber"] - 1)) + shifts["ShiftStart"] 
        shifts["ShiftEnd"]   = (1200 * (shifts["PeriodNumber"] - 1)) + shifts["ShiftEnd"] 
        
        # Insert game id and team
        shifts.insert(0, "GameId", self.game_id)   
        shifts.insert(2, "TeamName", self.home_team if home else self.away_team)   

        # Map team name tri-code to full name
        shifts["TeamName"] = shifts["TeamName"].replace(self._map_team_tri_to_names())
        
        shifts.drop("PlayerNr", axis=1, inplace=True)
        
        return shifts
    


import requests
import pandas as pd
import numpy as np
from pandera.typing import DataFrame


class GameJSON:
    """
    Extract information from the JSON report of an NHL game.
    
    Arguments:
        game_id: int, the id of the game, given as e.g. 2021020001.
        requests_session: optional, the session to use for scraping.
    
    Attributes:
        game_id: int, the id of the game.
        url: str, the url of the JSON report. One for play-by-play, and one for shifts.
        pbp: the play by play data.
        dressed_for_game: all players listed in the play by play data.
        shifts: all shifts reported for the game.
    
    Methods:
        get_event_data: retrieves the event data from the game.
        get_players_dressed_for_game: retrieves all players dressed for the game.
        get_shifts: retrieves all shifts from the home and away team.
    
    """
    def __init__(self, game_id: int, requests_session: requests.sessions.Session=None):
        """ Initialize a new game object. """
        
        # Create a new session if none is provided
        if requests_session is None:
            requests_session = requests.Session()
        
        # Save the game id as attribute
        self.game_id = game_id

        # Save the urls
        self.pbp_url = self.__get_url(game_id, report_type="PL")
        self.shifts_url = self.__get_url(game_id, report_type="SH")

        # Run the methods to save the results
        self.pbp = self.__get_event_data(requests_session)
        self.dressed_for_game = self.__get_players_dressed_for_game(requests_session)
        self.shifts = self.__get_game_shifts(requests_session)
        
    def __get_url(self, game_id: int, report_type: str="PL") -> str:
        """ Get the url used for retrieving data. """
        
        if report_type not in ["PL", "SH"]:
            raise ValueError("The report type is not supported.")
        
        # Specify the url
        if report_type == "PL":
            url = f"https://statsapi.web.nhl.com/api/v1/game/{game_id}/feed/live"
        elif report_type == "SH":
            url = f"https://api.nhle.com/stats/rest/en/shiftcharts?cayenneExp=gameId={game_id}"
            
        return url
    
            
    def __get_event_data(self, requests_session) -> DataFrame:
        """ Get play by play data from the NHL API for a given game. """
    
        # Get the json data from the game
        json_data = requests_session.get(self.pbp_url).json()
            
        # Get both team names
        team_name = [json_data["gameData"]["teams"]["away"]["name"],
                     json_data["gameData"]["teams"]["home"]["name"]]
        
        # Extract the play by play data
        pbp_data = pd.json_normalize(json_data["liveData"]["plays"]["allPlays"])
        
        # Add Game id
        pbp_data["GameId"] = self.game_id
        
        # Add team names
        pbp_data["AwayTeamName"] = team_name[0]
        pbp_data["HomeTeamName"] = team_name[1]
        
        # Specify columns to keep
        cols = ["GameId", "AwayTeamName", "HomeTeamName", "about.eventIdx",
                "about.period", "about.periodTime", "result.eventTypeId", "team.name",
                "about.goals.away", "about.goals.home", "coordinates.x", "coordinates.y",
                "players", "result.strength.name", "result.secondaryType",
                "result.gameWinningGoal", "result.emptyNet", "result.penaltySeverity",
                "result.penaltyMinutes"]
        
        # In case of missing columns
        if "result.emptyNet" not in pbp_data.columns:
            pbp_data["result.emptyNet"] = np.nan
        if "result.gameWinningGoal" not in pbp_data.columns:
            pbp_data["result.gameWinningGoal"] = np.nan
        if "result.penaltySeverity" not in pbp_data.columns:
            pbp_data["result.penaltySeverity"] = np.nan
        if "result.penaltyMinutes" not in pbp_data.columns:
            pbp_data["result.penaltyMinutes"] = np.nan
        
        # Specify renaming of columns
        column_renaming = {"about.eventIdx": "EventNumber", 
                           "about.period": "PeriodNumber",
                           "about.periodTime": "EventTime",
                           "result.eventTypeId": "EventType",
                           "team.name": "Team",
                           "about.goals.away": "GoalsAgainst",
                           "about.goals.home": "GoalsFor",
                           "coordinates.x": "X",
                           "coordinates.y": "Y",
                           "result.strength.name": "ScoringManpower",
                           "result.secondaryType": "Type",
                           "result.gameWinningGoal": "GameWinningGoal", 
                           "result.emptyNet": "EmptyNet", 
                           "result.penaltySeverity": "PenaltyType",
                           "result.penaltyMinutes": "PenaltyMinutes"}
        
        # If there is no data
        if len(pbp_data) == 0:
            return pd.DataFrame(columns=[
                "GameId", "AwayTeamName", "HomeTeamName", "EventNumber", "PeriodNumber",
                "EventTime", "TotalElapsedTime", "EventType", "Team", "GoalsAgainst", "GoalsFor", 
                "X", "Y", "ScoringManpower", "Type", "GameWinningGoal", "EmptyNet", "PenaltyType",
                "PenaltyMinutes", "PlayerType1",  "Player1", "PlayerId1", 
                "PlayerType2", "Player2", "PlayerId2", "PlayerType3", "Player3", "PlayerId3", 
                "AwayPlayer1", "AwayPlayerId1", "AwayPlayer2", "AwayPlayerId2", 
                "AwayPlayer3", "AwayPlayerId3", "AwayPlayer4", "AwayPlayerId4", 
                "AwayPlayer5", "AwayPlayerId5", "AwayPlayer6", "AwayPlayerId6", 
                "HomePlayer1", "HomePlayerId1", "HomePlayer2", "HomePlayerId2", 
                "HomePlayer3", "HomePlayerId3", "HomePlayer4", "HomePlayerId4", 
                "HomePlayer5", "HomePlayerId5", "HomePlayer6", "HomePlayerId6"])
        
        # Select columns
        pbp_data = pbp_data[cols]
        
        # Extract the players involved in the actions
        pbp_data["Player1"] = pbp_data["players"].apply(
            lambda x: x[0]["player"]["id"] if type(x) != float else x)
        
        pbp_data["Player2"] = pbp_data["players"].apply(
            lambda x: x[1]["player"]["id"] if type(x) != float and len(x) > 1 else np.nan)
        
        pbp_data["Player3"] = pbp_data["players"].apply(
            lambda x: x[2]["player"]["id"] if type(x) != float and len(x) > 2 else np.nan)

        # Extract player roles
        pbp_data["PlayerType1"] = pbp_data["players"].apply(
            lambda x: x[0]["playerType"] if type(x) != float else x)
        
        pbp_data["PlayerType2"] = pbp_data["players"].apply(
            lambda x: x[1]["playerType"] if type(x) != float and len(x) > 1 else np.nan)
        
        pbp_data["PlayerType3"] = pbp_data["players"].apply(
            lambda x: x[2]["playerType"] if type(x) != float and len(x) > 2 else np.nan)
        
        # Remove player column
        pbp_data.drop("players", axis=1, inplace=True)
        
        # Remove "goalie" role
        pbp_data.loc[pbp_data.PlayerType1.eq("Goalie"), "Player1"] = np.nan
        pbp_data.loc[pbp_data.PlayerType1.eq("Goalie"), "PlayerType1"] = np.nan
        pbp_data.loc[pbp_data.PlayerType2.eq("Goalie"), "Player2"] = np.nan
        pbp_data.loc[pbp_data.PlayerType2.eq("Goalie"), "PlayerType2"] = np.nan
        pbp_data.loc[pbp_data.PlayerType3.eq("Goalie"), "Player3"] = np.nan
        pbp_data.loc[pbp_data.PlayerType3.eq("Goalie"), "PlayerType3"] = np.nan
        
        # Rename columns
        pbp_data.rename(columns=column_renaming, inplace=True)
        
        # Add total elapsed time
        pbp_data["TotalElapsedTime"] = 1200 * (pbp_data["PeriodNumber"] - 1) + [
            60 * int(x[0]) + int(x[1])  for x in pbp_data["EventTime"].str.split(":")
            ]
        
        # Remove events not needed
        pbp_data = pbp_data.loc[~pbp_data.EventType.isin(["GAME_SCHEDULED", "PERIOD_OFFICIAL",
                                                          "PERIOD_READY"])]
        
        # Replace to match existing codebase
        pbp_data["EventType"] = pbp_data["EventType"].str.replace("_", " ")
        pbp_data["EventType"] = pbp_data["EventType"].str.replace("STOP", "STOPPAGE")
        
        return pbp_data
    
    
    def __get_players_dressed_for_game(self, requests_session) -> DataFrame:
        """ Get the players dressed for a given game. """

        # Get the json data from the game
        json_data = requests_session.get(self.pbp_url).json()
        
        # Get all players
        away_players = json_data["liveData"]["boxscore"]["teams"]["away"]["players"]
        home_players = json_data["liveData"]["boxscore"]["teams"]["home"]["players"]
        
        # Get the team names
        teams = [json_data["gameData"]["teams"]["away"]["name"],
                 json_data["gameData"]["teams"]["home"]["name"]]
        
        # Create a player id -> player position mapping for each team
        away_positions = {player["person"]["id"]: player["position"]["code"] for 
                          idx_str, player in away_players.items()}
        
        home_positions = {player["person"]["id"]: player["position"]["code"] for 
                          idx_str, player in home_players.items()}
        
        # Create a player id -> player name mapping for each team
        away_players = {player["person"]["id"]: player["person"]["fullName"] for 
                        idx_str, player in away_players.items()}
        
        home_players = {player["person"]["id"]: player["person"]["fullName"] for 
                        idx_str, player in home_players.items()}
        
        # Add team name
        away_players_with_team = {teams[0]: away_players}
        home_players_with_team = {teams[1]: home_players}
    
        # Combine players from both team into one dictionary
        players_with_team = {**away_players_with_team, **home_players_with_team}
        
        # Create a wide representation of the player and id mapping
        players_with_team_wide = pd.json_normalize(players_with_team.values())
        
        # Set a new column for team names
        players_with_team_wide["TeamName"] = teams
        players_with_team_wide["Side"] = ["Away", "Home"]
        
        # Create a data frame representation of the id name mapping
        game_players = players_with_team_wide.melt(id_vars=["TeamName", "Side"], 
                                                   var_name="PlayerId", 
                                                   value_name="Player").dropna(
                                                       ).reset_index(drop=True)
            
        # Add playing position
        game_players["Position"] = game_players["PlayerId"].replace({**away_positions, **home_positions})
        
        # Add a column for game id
        game_players.insert(0, column="GameId", value=self.game_id) 
        
        return game_players
    
    
    def __get_game_shifts(self, requests_session) -> DataFrame:
        """ Extract shift information from NHL's API for a given game. """
        
        # Extract shift information from NHL api
        game_shifts = requests_session.get(self.shifts_url).json()["data"]
        
        # Convert json to pandas data frame
        game_shifts_df = pd.json_normalize(game_shifts)
        
        # Add player name
        game_shifts_df["player"] = game_shifts_df[["firstName", "lastName"]].apply(
            lambda x: f"{x.firstName} {x.lastName}", axis=1)
        
        # Rename columns
        game_shifts_df.rename(columns={"startTime": "shiftStart", "endTime": "shiftEnd",
                                       "period": "periodNumber"}, inplace=True)
        
        # If there is no data for a given game
        if len(game_shifts_df) == 0:
            return pd.DataFrame(columns=["GameId", "PlayerId", "Player", "TeamName", 
                                         "ShiftNumber", "PeriodNumber", 
                                         "ShiftStart", "ShiftEnd", "Duration"])
        
        # Select the desired columns
        game_shifts_df = game_shifts_df[["gameId", "playerId", "player", "teamName", 
                                         "shiftNumber", "periodNumber", 
                                         "shiftStart", "shiftEnd", "duration"]]
        
        # Remove goals from the data
        game_shifts_df.dropna(inplace=True)
        
        # If there is no data (besides goals) for a given game
        if len(game_shifts_df) == 0:
            return pd.DataFrame(columns=["GameId", "PlayerId", "Player", "TeamName", 
                                         "ShiftNumber", "PeriodNumber", 
                                         "ShiftStart", "ShiftEnd", "Duration"])
        
        # In case of empty values
        missing_start_time = game_shifts_df["shiftStart"].eq("")
        missing_end_time = game_shifts_df["shiftEnd"].eq("")
        
        # Final period in regulation
        final_period = game_shifts_df["periodNumber"].eq(3)
        
        # Final two minutes of the period
        final_two_minutes = game_shifts_df["shiftStart"].apply(lambda x: int(x[0:2]) >= 18)
        
        # Update end time for final shifts in regulation
        game_shifts_df.loc[missing_end_time & final_period &
                           final_two_minutes, "shiftEnd"] = "20:00"
        
        # Update in case of new end times
        missing_end_time = game_shifts_df["shiftEnd"].eq("")
        
        # Compute start time for missing values
        game_shifts_df.loc[missing_start_time, "shiftStart"] = game_shifts_df.loc[missing_start_time, ["shiftEnd", "duration"]].apply(
            lambda x: f"{int(x.shiftEnd[:2]) - int(x.duration[:2])}:{int(x.shiftEnd[3:]) - int(x.duration[3:])}",
            axis=1)  
        
        # Compute end time for missing values
        game_shifts_df.loc[missing_end_time, "shiftEnd"] = game_shifts_df.loc[missing_end_time, ["shiftStart", "duration"]].apply(
            lambda x: f"{int(x.shiftStart[:2]) + int(x.duration[:2])}:{int(x.shiftStart[3:]) + int(x.duration[3:])}",
            axis=1)  
        
        # Fill any missing values for duration
        game_shifts_df["duration"] = game_shifts_df["duration"].fillna("00:00")
        
        # Convert time columns to seconds
        game_shifts_df["shiftStart"] = [int(minute) * 60 + int(second) for minute, second in 
                                        game_shifts_df["shiftStart"].str.split(':')]
        game_shifts_df["shiftEnd"]   = [int(minute) * 60 + int(second) for minute, second in 
                                        game_shifts_df["shiftEnd"].str.split(':')]
        game_shifts_df["duration"]  = [int(minute) * 60 + int(second) for minute, second in 
                                       game_shifts_df["duration"].str.split(':')]
        
        # Combine time with period number to get TotalElapsedTime
        game_shifts_df["shiftStart"] = 1200 * (game_shifts_df["periodNumber"] - 1) + game_shifts_df["shiftStart"]
        game_shifts_df["shiftEnd"]   = 1200 * (game_shifts_df["periodNumber"] - 1) + game_shifts_df["shiftEnd"]

        # For end times in separate periods
        period_shift = game_shifts_df.apply(lambda row: row["shiftEnd"] < row["shiftStart"], 
                                            axis=1)
        
        # Add one period worth of seconds
        game_shifts_df.loc[period_shift, "shiftEnd"] += 1200

        # Compute duration to ensure correctness to the highest level
        game_shifts_df["duration"] = game_shifts_df["shiftEnd"] - game_shifts_df["shiftStart"]

        # Rename columns
        game_shifts_df.columns = (game_shifts_df.columns.str.get(0).str.capitalize() + 
                                  game_shifts_df.columns.str.replace("^.{1}", "", regex=True))

        return game_shifts_df
        
        
    def get_event_data(self):
        """ Return the play-by-play data. """
        return self.pbp
    
    def get_players_dressed_for_game(self):
        """ Return the players dressed for the game. """
        return self.dressed_for_game
    
    def get_shifts(self):
        """ Return the shifts in the game. """
        return self.shifts


from hockey_api.nhl.html_scraper import GameHTML
from hockey_api.nhl.json_scraper import GameJSON
from hockey_api.nhl.merge_data import merge_json_pbp_and_shifts, merge_json_and_html_pbp, merge_json_and_html_shifts
from hockey_api.nhl.schedule_scraper import get_number_of_games_in_season, get_game_ids_between_dates
from hockey_api.nhl.utils import map_player_names_to_ids, map_player_ids_to_names, standardize_coordinates, \
    add_player_ids_to_html_shifts, adjust_player_roles_html
import pandas as pd
import requests
from tqdm import tqdm, trange
from time import sleep
from json import JSONDecodeError
from typing import List, Tuple, Union
from pandera.typing import DataFrame
from requests.exceptions import ConnectionError


def scrape_game(game_id: Union[str, int]) -> Tuple[DataFrame, DataFrame]:
    """
    Scrape play by play and shift data from a given game.

    Parameters
    ----------
    game_id : Union[str, int]
        The ID of the game, given as e.g. 2021020001 or "2021020001".

    Returns
    -------
    merged_game : DataFrame
        Combined information from the HTML and JSON representations of the game.
    merged_shifts : DataFrame
        Combined information from the HTML and JSON representations of the shifts.
        
    """
    # Specify a requests session
    requests_session = requests.Session()
    
    # Counter of failed scraping
    failed = 0
    
    try:
        # Get the JSON data
        json_game = GameJSON(game_id, requests_session)
    except KeyError:
        failed += 1
    
    # Merge json play by play and shifts
    try:
        json_game.pbp = merge_json_pbp_and_shifts(json_game)
    except (IndexError, ValueError, NameError, UnboundLocalError):
        if failed == 0:
            print(f"No JSON data/shifts for game {game_id} to merge. Only using HTML data.")
        failed += 1
    
    try:
        # Get the HTML data
        html_game = GameHTML(game_id, requests_session)
    except ValueError:
        failed += 1
    
    # If both the HTML and JSON retrievals failed
    if failed >= 2:
        print(f"The game id {game_id} provided does not exist, returning None.")
        return None

    # Add player names/ids to the data sources
    try:
        json_game.pbp = map_player_ids_to_names(json_game)
    except (IndexError, ValueError):
        print(f"Game {game_id} has issues with players on the ice in the JSON data.")

    html_game.pbp = map_player_names_to_ids(json_game, html_game)

    # Adjust player roles in the HTML data
    html_game.pbp = adjust_player_roles_html(html_game)

    # Merge JSON and HTML data
    merged_game = merge_json_and_html_pbp(json_game, html_game)
    
    # Standardize coordinates
    merged_game = standardize_coordinates(merged_game)
  
    # Get the JSON and HTML shift data
    json_shifts = json_game.get_shifts()
    html_shifts = pd.concat([html_game.get_shifts(home=True),
                             html_game.get_shifts(home=False)])
    
    # Add the player id column to the HTML shifts
    html_shifts = add_player_ids_to_html_shifts(json_shifts, html_shifts)
        
    # If there are JSON shifts
    if len(json_shifts) > 0:
        # Remove any player ids without a match, i.e. NaN
        html_shifts.dropna(subset=["PlayerId"], inplace=True)    
                
        # Merge the shifts from both sources
        merged_shifts = merge_json_and_html_shifts(json_shifts, html_shifts)
    else:
        # Placeholder value
        merged_shifts = html_shifts.fillna(-1)
    
    return merged_game, merged_shifts


def scrape_list_of_games(game_id_list: List) -> Tuple[DataFrame, DataFrame]:
    """
    Scrape all games in the provided list.

    Parameters
    ----------
    game_id_list : List
        List of all game ids to scrape.
   
    Returns
    -------
    games_in_list : DataFrame
        All events from the list of games, concatenated into one data frame.
    shifts_in_list : DataFrame
        All shifts from the list of games, concatenated into one data frame.

    """
    # Storage of all games
    game_dict = {}
    shift_dict = {}
    
    # Loop over all game ids 
    for game_id in tqdm(game_id_list, desc="Scraping list of games"):
        ## Default values
        # Return an empty data frame if scraping failed
        game = pd.DataFrame(columns=[
            "GameId", "Date", "AwayTeamName", "HomeTeamName", "EventNumber",
            "PeriodNumber", "Manpower", "EventTime", "TotalElapsedTime",
            "EventType", "Description", "Team", "GoalsAgainst", "GoalsFor", "X",
            "Y", "GameWinningGoal", "EmptyNet", "Zone", "ShotType", "PenaltyShot",
            "PenaltyType", "PenaltyMinutes", "PlayerType1", "Player1", "PlayerId1",
            "PlayerType2", "Player2", "PlayerId2", "PlayerType3", "Player3",
            "PlayerId3", "AwayPlayer1", "AwayPlayerId1", "AwayPlayer2",
            "AwayPlayerId2", "AwayPlayer3", "AwayPlayerId3", "AwayPlayer4",
            "AwayPlayerId4", "AwayPlayer5", "AwayPlayerId5", "AwayPlayer6",
            "AwayPlayerId6", "AwayGoalieOnIce", "AwayGoalieName", "AwayGoalieId",
            "HomePlayer1", "HomePlayerId1", "HomePlayer2", "HomePlayerId2",
            "HomePlayer3", "HomePlayerId3", "HomePlayer4", "HomePlayerId4",
            "HomePlayer5", "HomePlayerId5", "HomePlayer6", "HomePlayerId6",
            "HomeGoalieOnIce", "HomeGoalieName", "HomeGoalieId"])
        
        shifts = pd.DataFrame(columns=[
            "GameId", "PlayerId", "Player", "TeamName", "ShiftNumber",
            "PeriodNumber", "ShiftStart", "ShiftEnd", "Duration", "Event",
            "Source"])
        
        try:
            # Scrape the game
            scraped_data = scrape_game(game_id)
            
            # Check if the scraping worked
            if scraped_data is not None:
                game, shifts = scraped_data
                
        except (IndexError, ValueError, KeyError): 
            print(f"Scraping failed for game: {game_id}. Returning empty data frame.")
                    
        except ConnectionError:
            print("A connection error occurred, waiting 5 minutes before trying again...")
            sleep(300)
            
            # Scrape the game
            scraped_data = scrape_game(game_id)
            
            # Check if the scraping worked
            if scraped_data is not None:
                game, shifts = scraped_data
            
        except JSONDecodeError:
            print("JSON error found, waiting 10 seconds before trying again.")
            sleep(10)
            
            # Scrape the game
            scraped_data = scrape_game(game_id)
            
            # Check if the scraping worked
            if scraped_data is not None:
                game, shifts = scraped_data
        
        # Save the result in the dictionary
        game_dict[game_id] = game
        shift_dict[game_id] = shifts
    
    # Combine all games into one data frame
    games_in_list = pd.concat(game_dict).reset_index(drop=True)
    shifts_in_list = pd.concat(shift_dict).reset_index(drop=True)

    return games_in_list, shifts_in_list


def scrape_date_range(start_date: str, end_date: str=None) -> Tuple[DataFrame, DataFrame]:
    """
    Scrape all games from a given date range.

    Parameters
    ----------
    start_date : str
        Date in YYYY-MM-DD format.
    end_date : str, optional
        Date in YYYY-MM-DD format. If not provided defaults to start_date.

    Returns
    -------
    games_in_date_range : DataFrame
        All events from the date range, concatenated into one data frame.
    shifts_in_date_range : DataFrame
        All shifts from the date range, concatenated into one data frame.

    """
    # If end date is not supplied
    if end_date is None:
        end_date = start_date
    
    # Get all game ids for the given date or date range
    game_ids = get_game_ids_between_dates(start_date, end_date)
    
    # Storage of all games
    game_dict = {}
    shift_dict = {}

    # Loop over all game ids 
    for game_id in tqdm(game_ids,
                        desc=f"Scraping games between {start_date} and {end_date}"):
        ## Default values
        # Return an empty data frame if scraping failed
        game = pd.DataFrame(columns=[
            "GameId", "Date", "AwayTeamName", "HomeTeamName", "EventNumber",
            "PeriodNumber", "Manpower", "EventTime", "TotalElapsedTime",
            "EventType", "Description", "Team", "GoalsAgainst", "GoalsFor", "X",
            "Y", "GameWinningGoal", "EmptyNet", "Zone", "ShotType", "PenaltyShot",
            "PenaltyType", "PenaltyMinutes", "PlayerType1", "Player1", "PlayerId1",
            "PlayerType2", "Player2", "PlayerId2", "PlayerType3", "Player3",
            "PlayerId3", "AwayPlayer1", "AwayPlayerId1", "AwayPlayer2",
            "AwayPlayerId2", "AwayPlayer3", "AwayPlayerId3", "AwayPlayer4",
            "AwayPlayerId4", "AwayPlayer5", "AwayPlayerId5", "AwayPlayer6",
            "AwayPlayerId6", "AwayGoalieOnIce", "AwayGoalieName", "AwayGoalieId",
            "HomePlayer1", "HomePlayerId1", "HomePlayer2", "HomePlayerId2",
            "HomePlayer3", "HomePlayerId3", "HomePlayer4", "HomePlayerId4",
            "HomePlayer5", "HomePlayerId5", "HomePlayer6", "HomePlayerId6",
            "HomeGoalieOnIce", "HomeGoalieName", "HomeGoalieId"])
        
        shifts = pd.DataFrame(columns=[
            "GameId", "PlayerId", "Player", "TeamName", "ShiftNumber",
            "PeriodNumber", "ShiftStart", "ShiftEnd", "Duration", "Event",
            "Source"])
        
        try:
            # Scrape the game
            scraped_data = scrape_game(game_id)
            
            # Check if the scraping worked
            if scraped_data is not None:
                game, shifts = scraped_data
        
        except (IndexError, ValueError, KeyError): 
            print(f"Scraping failed for game: {game_id}. Returning empty data frame.")
            
        except ConnectionError:
            print("A connection error occurred, waiting 5 minutes before trying again...")
            sleep(300)
            
            # Scrape the game
            scraped_data = scrape_game(game_id)
            
            # Check if the scraping worked
            if scraped_data is not None:
                game, shifts = scraped_data
            
        except JSONDecodeError:
            print("JSON error found, waiting 10 seconds before trying again.")
            sleep(10)
            
            # Scrape the game
            scraped_data = scrape_game(game_id)
            
            # Check if the scraping worked
            if scraped_data is not None:
                game, shifts = scraped_data
            
        # Save the result in the dictionary
        game_dict[game_id] = game
        shift_dict[game_id] = shifts
        
    # Combine all games into one data frame
    games_in_date_range = pd.concat(game_dict).reset_index(drop=True)
    shifts_in_date_range = pd.concat(shift_dict).reset_index(drop=True)
    
    return games_in_date_range, shifts_in_date_range


def scrape_season(season: Union[str, int], season_type: str="R") -> Tuple[DataFrame, DataFrame]:
    """
    Scrape all games and shifts from a given part of a season.

    Parameters
    ----------
    season : Union[str, int]
        The starting year of the season to consider, e.g. 2021 for 2021-2022.
    season_type : str
        One of: R = Regular Season, P = Playoffs, PR = Preseason, A = All-star.

    Returns
    -------
    games_in_season : DataFrame
        All events from the season, concatenated into one data frame.
    shifts_in_season : DataFrame
        All shifts from the season, concatenated into one data frame.

    """
    # Get the number of games in the season and the season type code
    nr_of_games, season_type_code = get_number_of_games_in_season(season, season_type)
    
    # Storage of all games
    game_dict = {}
    shift_dict = {}

    # The base for game id
    game_id_base = f"{season}{season_type_code}"
    
    # Loop over all game ids 
    for game_id_nr in trange(1, nr_of_games+1, 
                             desc=f"Scraping season {season}-{int(season)+1}"):
        
        # The padding to be added with zeros
        padding = ""
        
        # Pad with zeros if needed
        if game_id_nr < 10:
            padding = "000"
        elif game_id_nr >= 10 and game_id_nr < 100:
            padding = "00"
        elif game_id_nr >= 100 and game_id_nr < 1000:
            padding = "0"
        
        # Get the game id
        game_id = f"{game_id_base}{padding}{game_id_nr}"
        
        ## Default values
        # Return an empty data frame if scraping failed
        game = pd.DataFrame(columns=[
            "GameId", "Date", "AwayTeamName", "HomeTeamName", "EventNumber",
            "PeriodNumber", "Manpower", "EventTime", "TotalElapsedTime",
            "EventType", "Description", "Team", "GoalsAgainst", "GoalsFor", "X",
            "Y", "GameWinningGoal", "EmptyNet", "Zone", "ShotType", "PenaltyShot",
            "PenaltyType", "PenaltyMinutes", "PlayerType1", "Player1", "PlayerId1",
            "PlayerType2", "Player2", "PlayerId2", "PlayerType3", "Player3",
            "PlayerId3", "AwayPlayer1", "AwayPlayerId1", "AwayPlayer2",
            "AwayPlayerId2", "AwayPlayer3", "AwayPlayerId3", "AwayPlayer4",
            "AwayPlayerId4", "AwayPlayer5", "AwayPlayerId5", "AwayPlayer6",
            "AwayPlayerId6", "AwayGoalieOnIce", "AwayGoalieName", "AwayGoalieId",
            "HomePlayer1", "HomePlayerId1", "HomePlayer2", "HomePlayerId2",
            "HomePlayer3", "HomePlayerId3", "HomePlayer4", "HomePlayerId4",
            "HomePlayer5", "HomePlayerId5", "HomePlayer6", "HomePlayerId6",
            "HomeGoalieOnIce", "HomeGoalieName", "HomeGoalieId"])
        
        shifts = pd.DataFrame(columns=[
            "GameId", "PlayerId", "Player", "TeamName", "ShiftNumber",
            "PeriodNumber", "ShiftStart", "ShiftEnd", "Duration", "Event",
            "Source"])
        
        try:
            # Scrape the game
            scraped_data = scrape_game(game_id)
            
            # Check if the scraping worked
            if scraped_data is not None:
                game, shifts = scraped_data
                
        except (IndexError, ValueError, KeyError): 
            print(f"Scraping failed for game: {game_id}. Returning empty data frame.")
        
        except ConnectionError:
            print("A connection error occurred, waiting 5 minutes before trying again...")
            sleep(300)
            
            # Scrape the game
            scraped_data = scrape_game(game_id)
            
            # Check if the scraping worked
            if scraped_data is not None:
                game, shifts = scraped_data            
            
        except JSONDecodeError:
            print("JSON error found, waiting 10 seconds before trying again.")
            sleep(10)
            
            # Scrape the game
            scraped_data = scrape_game(game_id)
            
            # Check if the scraping worked
            if scraped_data is not None:
                game, shifts = scraped_data
        
        # Save the result in the dictionary
        game_dict[game_id] = game
        shift_dict[game_id] = shifts
        
    # Combine all games into one data frame
    games_in_season = pd.concat(game_dict).reset_index(drop=True)
    shifts_in_season = pd.concat(shift_dict).reset_index(drop=True)

    return games_in_season, shifts_in_season



import requests
import re
import numpy as np
import pandas as pd
from difflib import get_close_matches
from itertools import chain
from pandera.typing import DataFrame


def get_player_data() -> DataFrame:
    """
    Download all player data from NHL.com.

    Returns
    -------
    players : DataFrame
        All information available about all players recorded NHL's database, 
        upto the date of retrieval.

    """
    # Get all players from the NHL api
    player_data = requests.get("https://records.nhl.com/site/api/player").json()

    # Covnert json file to data frame
    players = pd.json_normalize(player_data["data"])

    return players


def adjust_player_roles_html(html_game) -> DataFrame:
    """
    Adjust the player roles in the HTML play by play data to match the correct order.

    Parameters
    ----------
    html_game : GameHTML
        Object containing the HTML representation of the game.

    Returns
    -------
    html_pbp : DataFrame
        Modified HTML play by play data frame adjusted player roles.

    """
    # Get the play by play data from the JSON object
    html_pbp = html_game.pbp.copy()

    # Find if the player involved was a home player
    home_player = html_pbp.apply(lambda x: x.PlayerId1 in [x.HomePlayerId1, x.HomePlayerId2, 
                                                           x.HomePlayerId3, x.HomePlayerId4, 
                                                           x.HomePlayerId5, x.HomePlayerId6], axis=1)
    # Find if the player involved was an away player
    away_player = html_pbp.apply(lambda x: x.PlayerId1 in [x.AwayPlayerId1, x.AwayPlayerId2, 
                                                           x.AwayPlayerId3, x.AwayPlayerId4, 
                                                           x.AwayPlayerId5, x.AwayPlayerId6], axis=1)
    
    # Determine which team had the event
    home_team_event = html_pbp.Team.eq(html_pbp.HomeTeamName)
    away_team_event = html_pbp.Team.eq(html_pbp.AwayTeamName)
    
    # Get the players that are in the wrong order
    home_player_1 = html_pbp.loc[away_team_event & home_player, ["PlayerId1", "Player1"]].values
    away_player_1 = html_pbp.loc[home_team_event & away_player, ["PlayerId1", "Player1"]].values

    home_player_2 = html_pbp.loc[away_team_event & home_player, ["PlayerId2", "Player2"]].values
    away_player_2 = html_pbp.loc[home_team_event & away_player, ["PlayerId2", "Player2"]].values
    
    # Update the player roles
    html_pbp.loc[away_team_event & home_player, ["PlayerId1", "Player1"]] = home_player_2
    html_pbp.loc[home_team_event & away_player, ["PlayerId1", "Player1"]] = away_player_2

    html_pbp.loc[away_team_event & home_player, ["PlayerId2", "Player2"]] = home_player_1
    html_pbp.loc[home_team_event & away_player, ["PlayerId2", "Player2"]] = away_player_1

    return html_pbp


def add_player_ids_to_html_shifts(json_shifts: DataFrame, html_shifts: DataFrame) -> DataFrame:
    """
    Add a player id column to the HTML shifts.

    Parameters
    ----------
    json_shifts : DataFrame
        All shifts from the game and its JSON representation.
    html_shifts : DataFrame
        All shifts from the game and its HTML representation.

    Returns
    -------
    html_shifts : DataFrame
        Modified HTML shifts with a new column for player id.

    """
    # Storage for player name and id mapping
    player_name_to_id_map = {}
    
    # Get all unique players from the HTML shifts
    html_players = html_shifts[["Player", "TeamName"]].drop_duplicates()
    
    # Get all unique players from the JSON shifts
    json_players = json_shifts[["Player", "PlayerId", "TeamName"]].drop_duplicates()
    
    # Loop over both teams
    for team, team_players in html_players.groupby("TeamName"):
        # Get all players from the current team
        json_team_players = json_players.loc[json_players.TeamName.eq(team)].copy()
        
        # Loop over each player in the team
        for player in team_players.Player:
            # Get the closest player from the list
            match = get_close_matches(player, json_team_players.Player.str.upper().to_list(), n=1)
        
            # If there was a match
            if len(match) > 0:
                # Find the index of the matched player name
                match_idx = json_team_players.Player.str.upper().eq(match[0])
                
                # Get the player id from the matched player
                match_player_id = json_team_players.loc[match_idx, "PlayerId"].values[0]
                
                # Remove the matched player from the possibilites
                json_team_players.drop(match_idx[match_idx].index, inplace=True)
                
            else:
                # If there is no match
                match_player_id = np.nan
                
            # Save the player id in the dictionary
            player_name_to_id_map[player] = match_player_id

    # Copy to avoid changing in-place
    html_shifts = html_shifts.copy()

    # Add player id as a column
    html_shifts.insert(1, "PlayerId", html_shifts.Player.replace(player_name_to_id_map))

    return html_shifts


def map_player_ids_to_names(json_game) -> DataFrame:
    """
    Add player names to the JSON data.

    Parameters
    ----------
    json_game : GameJSON
        Object containing the JSON representation of the game.

    Returns
    -------
    pbp : DataFrame
        Modified JSON play by play data frame with player names added.

    """
    # Get the play by play data from the JSON object
    json_pbp = json_game.pbp.copy()
     
    # Get all players dressed for the game
    players_in_game = json_game.get_players_dressed_for_game()
    
    # Extract the column names of home and away players
    home_players = json_pbp.columns.str.extract("(HomePlayer[1-6])").dropna().values[:, 0]
    away_players = json_pbp.columns.str.extract("(AwayPlayer[1-6])").dropna().values[:, 0]
    
    # Add a new column for player ids
    json_pbp[[re.sub("Player", "PlayerId", i) for i in home_players]] = json_pbp[home_players]
    json_pbp[[re.sub("Player", "PlayerId", i) for i in away_players]] = json_pbp[away_players]
    
    # Map player id to player name
    home_player_map = players_in_game.loc[players_in_game.Side.eq("Home")].set_index("PlayerId")["Player"].to_dict()
    away_player_map = players_in_game.loc[players_in_game.Side.eq("Away")].set_index("PlayerId")["Player"].to_dict()

    # Remap away player id columns to away player names
    json_pbp[away_players] = json_pbp[away_players].replace(
        away_player_map)
    
    # Remap home player id columns to home player names
    json_pbp[home_players] = json_pbp[home_players].replace(
        home_player_map)
    
    # Add new columns for player involvement and map to player names
    for i in range(1, 4):
        json_pbp[f"PlayerId{i}"] = json_pbp[f"Player{i}"]
        
        json_pbp[f"Player{i}"] = json_pbp[f"Player{i}"].replace(
            {**away_player_map, **home_player_map})
    
    # Create a (nested) list of all player role columns
    player_role_columns = [json_pbp.columns[json_pbp.columns.str.contains(
        f"^Player{i}$|^PlayerType{i}")].sort_values(ascending=False).union(
        json_pbp.columns[json_pbp.columns.str.contains(f"^PlayerId{i}")], 
        sort=False).tolist() for i in range(1, 4)]
    
    # Unnest the list of lists into a list
    player_role_columns = list(chain.from_iterable(player_role_columns))
    
    # Create a (nested) list of all away players
    away_player_columns = [json_pbp.columns[json_pbp.columns.str.contains(
        f"AwayPlayer{i}")].sort_values(ascending=False).union(
        json_pbp.columns[json_pbp.columns.str.contains(f"AwayPlayerId{i}")], 
        sort=False).tolist() for i in range(1, 7)]
    
    # Unnest the list of lists into a list        
    away_player_columns = list(chain.from_iterable(away_player_columns))
    
    # Create a (nested) list of all home players
    home_player_columns = [json_pbp.columns[json_pbp.columns.str.contains(
        f"HomePlayer{i}")].sort_values(ascending=False).union(
        json_pbp.columns[json_pbp.columns.str.contains(f"HomePlayerId{i}")], 
        sort=False).tolist() for i in range(1, 7)]
    
    # Unnest the list of lists into a list        
    home_player_columns = list(chain.from_iterable(home_player_columns))
    
    # Reorder columns
    json_pbp = json_pbp[
        ["GameId", "AwayTeamName", "HomeTeamName", "EventNumber", "PeriodNumber",
         "EventTime", "TotalElapsedTime", "EventType", "Team", "GoalsAgainst", "GoalsFor", 
         "X", "Y", "ScoringManpower", "Type", "GameWinningGoal", "EmptyNet", "PenaltyType",
         "PenaltyMinutes"] + player_role_columns + away_player_columns + home_player_columns]
    
    return json_pbp


def map_player_names_to_ids(json_game, html_game) -> DataFrame:
    """
    Add player ids to the HMTL data.

    Parameters
    ----------
    json_game : GameJSON
        Object containing the JSON representation of the game.
    html_game : GameHTML
        Object containing the HTML representation of the game.

    Returns
    -------
    html_pbp : DataFrame
        Modified HTML play by play data frame with player names added.

    """
    # Get the play by play data form the HTML report
    html_pbp = html_game.get_event_data().copy()
    
    # Get the players dressed for the game
    players_in_game = json_game.get_players_dressed_for_game()
    
    # Get the players from the shifts in the JSON representation
    json_shifts = json_game.get_shifts().drop_duplicates("PlayerId")
    
    # Combine players dressed for the game and shifts
    all_players = players_in_game.merge(json_shifts[["GameId", "PlayerId", "Player", "TeamName"]],
                                        how="outer")
    
    # Do a self-join to fill values for the side
    all_players_unique = all_players.drop("Side", axis=1).merge(
        all_players[["TeamName", "Side"]].drop_duplicates(), 
        on="TeamName", how="outer").dropna(subset=["Side"])
    
    # Fill NA values for position with N/A to match dressed for game structure
    all_players_unique["Position"] = all_players_unique["Position"].fillna("N/A")
    
    # Reset the index and overwrite players in game
    players_in_game = all_players_unique.reset_index(drop=True)    
    
    # Extract the column names of home and away players
    home_players = html_pbp.columns.str.extract("(HomePlayer[1-6])").dropna().values[:, 0]
    away_players = html_pbp.columns.str.extract("(AwayPlayer[1-6])").dropna().values[:, 0]
    
    # Column names for id columns
    home_player_ids = [re.sub("Player", "PlayerId", i) for i in home_players]
    away_player_ids = [re.sub("Player", "PlayerId", i) for i in away_players]
    
    # Add a new column for player ids
    html_pbp[[re.sub("Player", "PlayerId", i) for i in home_players]] = html_pbp[home_players]
    html_pbp[[re.sub("Player", "PlayerId", i) for i in away_players]] = html_pbp[away_players]
    
    # Map player id to player name
    home_player_map = players_in_game.loc[players_in_game.Side.eq("Home")].set_index("Player")["PlayerId"].to_dict()
    away_player_map = players_in_game.loc[players_in_game.Side.eq("Away")].set_index("Player")["PlayerId"].to_dict()

    # Convert dictionary keys to all caps to match HTML format
    home_player_map = {player.upper(): player_id for player, player_id in home_player_map.items()}
    away_player_map = {player.upper(): player_id for player, player_id in away_player_map.items()}

    # Remap home player id columns to home player names
    html_pbp[home_player_ids] = html_pbp[home_player_ids].replace(
        home_player_map)
    
    # Remap away player id columns to away player names
    html_pbp[away_player_ids] = html_pbp[away_player_ids].replace(
        away_player_map)
    
    # Fix any possible unmatched names
    unmatched_home = any(html_pbp[home_player_ids].dtypes == "object")
    unmatched_away = any(html_pbp[away_player_ids].dtypes == "object")
    
    if unmatched_home:
        # Find all home players
        unique_home_players = html_pbp[home_player_ids].melt().value.dropna().unique()
        
        # Find the unmapped home players
        unmapped_home_players = [player for player in unique_home_players if isinstance(player, str)]
        
        # Find all the player ids that have been matched successfully
        mapped_home_player_ids = set(unique_home_players).symmetric_difference(unmapped_home_players)
        
        # Find all players who were not matched
        unmapped_home_player_ids = {player: player_id for player, player_id in home_player_map.items() if
                                    player_id not in mapped_home_player_ids}
        
        # Loop through all unmapped players
        for unmapped_home_player in unmapped_home_players:
            try:
                # Get the closest match from the unmapped players
                unmapped = get_close_matches(unmapped_home_player, unmapped_home_player_ids, n=1)[0]
                
                # Save the name in the dictionary
                home_player_map[unmapped_home_player] = home_player_map[unmapped]
                
                # Delete the other name
                del home_player_map[unmapped]
        
            except IndexError:
                print(f"No player match found for {unmapped_home_player}, mapping -1.")
                # Default value
                home_player_map[unmapped_home_player] = -1
                
        # Remap home player id columns to home player names
        html_pbp[home_player_ids] = html_pbp[home_player_ids].replace(
            home_player_map)
        
    if unmatched_away:
        # Find all away players
        unique_away_players = html_pbp[away_player_ids].melt().value.dropna().unique()
        
        # Find the unmapped away players
        unmapped_away_players = [player for player in unique_away_players if isinstance(player, str)]
        
        # Find all the player ids that have been matched successfully
        mapped_away_player_ids = set(unique_away_players).symmetric_difference(unmapped_away_players)
        
        # Find all players who were not matched
        unmapped_away_player_ids = {player: player_id for player, player_id in away_player_map.items() if
                                    player_id not in mapped_away_player_ids}
 
        # Loop through all unmapped players
        for unmapped_away_player in unmapped_away_players:
            try:
                # Get the closest match from the unmapped players
                unmapped = get_close_matches(unmapped_away_player, unmapped_away_player_ids, n=1)[0]
                
                # Save the name in the dictionary
                away_player_map[unmapped_away_player] = away_player_map[unmapped]
                
                # Delete the other name
                del away_player_map[unmapped]
        
            except IndexError:
                print(f"No player match found for {unmapped_away_player}, mapping -1.")
                # Default value
                away_player_map[unmapped_away_player] = -1
                
        # Remap away player id columns to away player names
        html_pbp[away_player_ids] = html_pbp[away_player_ids].replace(
            away_player_map)
    
    # Find all unique player names among the roles
    player_names = pd.Series(html_pbp[["Player1", "Player2", "Player3"]].melt().value.dropna().unique())
    
    # Combine away and home perspective
    player_maps = {**away_player_map, **home_player_map}
    
    # Check if there are any missing player names
    missing_maps = ~player_names.isin(player_maps)
    
    if any(missing_maps):
        # Name of the player(s) not mapped
        keys = player_names[missing_maps].values
        
        # Set each missing map to nan
        for key in keys:
            player_maps[key] = np.nan 
    
    # Add new columns for player involvement and map to player names
    for i in range(1, 4):
        # Initialize new column
        html_pbp[f"PlayerId{i}"] = html_pbp[f"Player{i}"]
        
        # Map names to ids
        player_ids = html_pbp[f"PlayerId{i}"].replace(player_maps)    
        
        # Save as new column in play by play data
        html_pbp[f"PlayerId{i}"] = player_ids
        
    # Map goalie names to ids
    html_pbp["HomeGoalieId"] = html_pbp["HomeGoalieName"].replace(home_player_map)
    html_pbp["AwayGoalieId"] = html_pbp["AwayGoalieName"].replace(away_player_map)
        
    # Reorder columns
    html_pbp = html_pbp[
        ["GameId", "Date", "AwayTeamName", "HomeTeamName", "EventNumber", 
         "PeriodNumber", "Manpower", "EventTime", "TotalElapsedTime",
         "EventType", "Description", "Team", "Zone", "ShotType", "PenaltyShot", 
         "PenaltyType", "PenaltyMinutes", "PlayerType1", "Player1", "PlayerId1", 
         "PlayerType2", "Player2", "PlayerId2", "PlayerType3", "Player3", "PlayerId3", 
         "AwayPlayer1", "AwayPlayerId1", "AwayPlayer2", "AwayPlayerId2", 
         "AwayPlayer3", "AwayPlayerId3", "AwayPlayer4", "AwayPlayerId4", 
         "AwayPlayer5", "AwayPlayerId5", "AwayPlayer6", "AwayPlayerId6",
         "AwayGoalieOnIce", "AwayGoalieName", "AwayGoalieId",
         "HomePlayer1", "HomePlayerId1", "HomePlayer2", "HomePlayerId2", 
         "HomePlayer3", "HomePlayerId3", "HomePlayer4", "HomePlayerId4", 
         "HomePlayer5", "HomePlayerId5", "HomePlayer6", "HomePlayerId6",
         "HomeGoalieOnIce", "HomeGoalieName", "HomeGoalieId"]]
    
    return html_pbp


def standardize_coordinates(pbp: DataFrame) -> DataFrame:
    """
    Synchronize the attacking direction within games such that both teams have
    their own net located at x=-89 and opposition's net at x=89.

    Parameters
    ----------
    pbp : DataFrame
        Data frame of play by play data.

    Returns
    -------
    pbp : DataFrame
        Adjusted data frame with new X and Y coordinates.

    """
    
    # Copy to avoid changing in-place
    pbp = pbp.copy()
    
    # Check if the team was the home team
    pbp["IsHome"] = pbp.Team.eq(pbp.HomeTeamName)
        
    # Check if the game was in overtime 
    pbp["Overtime"] = pbp.PeriodNumber.ge(4)
    
    # Get all shot events in regulation where the goalie is on the ice
    shot_events = pbp.loc[pbp.EventType.isin(["GOAL", "MISSED SHOT", "SHOT"]) &
                          ((pbp.IsHome & pbp.AwayGoalieOnIce) | 
                           (~pbp.IsHome & pbp.HomeGoalieOnIce))]
                                  
    # Compute the game median x-coordinate and number of shots by group
    game_median_x = shot_events.groupby(["GameId", "IsHome", "PeriodNumber", "Overtime"], 
                                        as_index=False).agg({"X": ["median", "size"]}
                                                            ).reset_index(drop=True)
    # Rename columns
    game_median_x.columns = ["GameId", "IsHome", "PeriodNumber", "Overtime", "Xmedian", "Size"]
    
    # Check if the median x coordinate was positive or negative
    game_median_x["Sign"] = np.sign(game_median_x["Xmedian"])
        
    # Add sign, indicating if the attack was toward 100 or -100
    pbp = pbp.merge(game_median_x[["GameId", "IsHome", "PeriodNumber", "Overtime", "Sign"]], 
                    how="left")
    
    # Adjust the X and Y coordinates by considering the sign
    pbp.insert(int(np.where(pbp.columns.str.contains("^X$"))[0][0]), 
               "X_adj", pbp["X"] * pbp["Sign"])
    pbp.insert(int(np.where(pbp.columns.str.contains("^Y$"))[0][0]), 
               "Y_adj", pbp["Y"] * pbp["Sign"])

    # Remove the unwanted columns
    pbp.drop(["IsHome", "Overtime", "Sign"], axis=1, inplace=True)    

    return pbp
